@@ -80,14 +80,11 @@ pub enum XlsbError {
     Password,
     /// Worksheet not found
     WorksheetNotFound(String),
-    /// XML Encoding error
-    Encoding(quick_xml::encoding::EncodingError),
 }
 
 from_err!(std::io::Error, XlsbError, Io);
 from_err!(zip::result::ZipError, XlsbError, Zip);
 from_err!(quick_xml::Error, XlsbError, Xml);
-from_err!(quick_xml::encoding::EncodingError, XlsbError, Encoding);
 
 impl std::fmt::Display for XlsbError {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
@@ -117,7 +114,6 @@ impl std::fmt::Display for XlsbError {
             }
             XlsbError::Password => write!(f, "Workbook is password protected"),
             XlsbError::WorksheetNotFound(name) => write!(f, "Worksheet '{name}' not found"),
-            XlsbError::Encoding(e) => write!(f, "XML encoding error: {e}"),
         }
     }
 }
@@ -163,11 +159,10 @@ impl<RS: Read + Seek> Xlsb<RS> {
         match self.zip.by_name("xl/_rels/workbook.bin.rels") {
             Ok(f) => {
                 let mut xml = XmlReader::from_reader(BufReader::new(f));
-                let config = xml.config_mut();
-                config.check_end_names = false;
-                config.trim_text(false);
-                config.check_comments = false;
-                config.expand_empty_elements = true;
+                xml.check_end_names(false)
+                    .trim_text(false)
+                    .check_comments(false)
+                    .expand_empty_elements(true);
                 let mut buf: Vec<u8> = Vec::with_capacity(64);
 
                 loop {
@@ -187,12 +182,7 @@ impl<RS: Read + Seek> Xlsb<RS> {
                                         key: QName(b"Target"),
                                         value: v,
                                     } => {
-                                        target = Some(
-                                            xml.decoder()
-                                                .decode(&v)
-                                                .map_err(XlsbError::Encoding)?
-                                                .into_owned(),
-                                        );
+                                        target = Some(xml.decoder().decode(&v)?.into_owned());
                                     }
                                     _ => (),
                                 }
