@@ -89,10 +89,6 @@ pub enum XlsxError {
     TableNotFound(String),
     /// The specified sheet is not a worksheet
     NotAWorksheet(String),
-    /// XML Encoding error
-    Encoding(quick_xml::encoding::EncodingError),
-    /// XML attribute error
-    XmlAttribute(quick_xml::events::attributes::AttrError),
 }
 
 from_err!(std::io::Error, XlsxError, Io);
@@ -102,12 +98,6 @@ from_err!(quick_xml::Error, XlsxError, Xml);
 from_err!(std::string::ParseError, XlsxError, Parse);
 from_err!(std::num::ParseFloatError, XlsxError, ParseFloat);
 from_err!(std::num::ParseIntError, XlsxError, ParseInt);
-from_err!(quick_xml::encoding::EncodingError, XlsxError, Encoding);
-from_err!(
-    quick_xml::events::attributes::AttrError,
-    XlsxError,
-    XmlAttribute
-);
 
 impl std::fmt::Display for XlsxError {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
@@ -149,8 +139,6 @@ impl std::fmt::Display for XlsxError {
             XlsxError::Password => write!(f, "Workbook is password protected"),
             XlsxError::TableNotFound(n) => write!(f, "Table '{n}' not found"),
             XlsxError::NotAWorksheet(typ) => write!(f, "Expecting a worksheet, got {typ}"),
-            XlsxError::Encoding(e) => write!(f, "XML encoding error: {e}"),
-            XlsxError::XmlAttribute(e) => write!(f, "XML attribute error: {e}"),
         }
     }
 }
@@ -165,8 +153,6 @@ impl std::error::Error for XlsxError {
             XlsxError::Parse(e) => Some(e),
             XlsxError::ParseInt(e) => Some(e),
             XlsxError::ParseFloat(e) => Some(e),
-            XlsxError::Encoding(e) => Some(e),
-            XlsxError::XmlAttribute(e) => Some(e),
             _ => None,
         }
     }
@@ -344,14 +330,13 @@ impl<RS: Read + Seek> Xlsx<RS> {
                                 key: QName(b"name"),
                                 ..
                             } => {
-                                name = a.decode_and_unescape_value(xml.decoder())?.to_string();
+                                name = a.decode_and_unescape_value(&xml)?.to_string();
                             }
                             Attribute {
                                 key: QName(b"state"),
                                 ..
                             } => {
-                                visible = match a.decode_and_unescape_value(xml.decoder())?.as_ref()
-                                {
+                                visible = match a.decode_and_unescape_value(&xml)?.as_ref() {
                                     "visible" => SheetVisible::Visible,
                                     "hidden" => SheetVisible::Hidden,
                                     "veryHidden" => SheetVisible::VeryHidden,
@@ -408,7 +393,7 @@ impl<RS: Read + Seek> Xlsx<RS> {
                 Ok(Event::Start(ref e)) if e.name().as_ref() == b"workbookPr" => {
                     self.is_1904 = match e.try_get_attribute("date1904")? {
                         Some(c) => ["1", "true"].contains(
-                            &c.decode_and_unescape_value(xml.decoder())
+                            &c.decode_and_unescape_value(&xml)
                                 .map_err(XlsxError::Xml)?
                                 .as_ref(),
                         ),
@@ -421,7 +406,7 @@ impl<RS: Read + Seek> Xlsx<RS> {
                         .filter_map(std::result::Result::ok)
                         .find(|a| a.key == QName(b"name"))
                     {
-                        let name = a.decode_and_unescape_value(xml.decoder())?.to_string();
+                        let name = a.decode_and_unescape_value(&xml)?.to_string();
                         val_buf.clear();
                         let mut value = String::new();
                         loop {
@@ -1103,11 +1088,10 @@ fn xml_reader<'a, RS: Read + Seek>(
     match zip.by_name(&actual_path) {
         Ok(f) => {
             let mut r = XmlReader::from_reader(BufReader::new(f));
-            let config = r.config_mut();
-            config.check_end_names = false;
-            config.trim_text(false);
-            config.check_comments = false;
-            config.expand_empty_elements = true;
+            r.check_end_names(false)
+                .trim_text(false)
+                .check_comments(false)
+                .expand_empty_elements(true);
             Some(Ok(r))
         }
         Err(ZipError::FileNotFound) => None,
@@ -1238,7 +1222,7 @@ fn get_row_and_optional_column(range: &[u8]) -> Result<(u32, Option<u32>), XlsxE
 /// attempts to read either a simple or richtext string
 pub(crate) fn read_string(
     xml: &mut XlReader<'_>,
-    closing: QName,
+    QName(closing): QName,
 ) -> Result<Option<String>, XlsxError> {
     let mut buf = Vec::with_capacity(1024);
     let mut val_buf = Vec::with_capacity(1024);
@@ -1256,7 +1240,7 @@ pub(crate) fn read_string(
             Ok(Event::Start(ref e)) if e.local_name().as_ref() == b"rPh" => {
                 is_phonetic_text = true;
             }
-            Ok(Event::End(ref e)) if e.name() == closing => {
+            Ok(Event::End(ref e)) if e.local_name().as_ref() == closing => {
                 return Ok(rich_buffer);
             }
             Ok(Event::End(ref e)) if e.local_name().as_ref() == b"rPh" => {
@@ -1277,7 +1261,7 @@ pub(crate) fn read_string(
                     s.push_str(&value);
                 } else {
                     // consume any remaining events up to expected closing tag
-                    xml.read_to_end_into(closing, &mut val_buf)?;
+                    xml.read_to_end_into(QName(closing), &mut val_buf)?;
                     return Ok(Some(value));
                 }
             }
@@ -1420,9 +1404,6 @@ pub(crate) fn coordinate_to_name(cell: (u32, u32)) -> Result<Vec<u8>, XlsxError>
 #[cfg(test)]
 mod tests {
     use super::*;
-    use std::io::Write;
-    use zip::write::SimpleFileOptions;
-    use zip::ZipWriter;
 
     #[test]
     fn test_dimensions() {
@@ -1516,59 +1497,4 @@ mod tests {
             "A2 is a cell, B2 is another, also C108, but XFE123 is not and \"A3\" in quote wont change.".to_owned()
         );
     }
-
-    #[test]
-    fn test_read_shared_strings_with_namespaced_si_name() {
-        let shared_strings_data = br#"<?xml version="1.0" encoding="utf-8"?>
-<x:sst count="1187" uniqueCount="1187" xmlns:x="http://schemas.openxmlformats.org/spreadsheetml/2006/main">
-    <x:si>
-        <x:t>String 1</x:t>
-    </x:si>
-    <x:si>
-        <x:r>
-            <x:rPr>
-                <x:sz val="11"/>
-            </x:rPr>
-            <x:t>String 2</x:t>
-        </x:r>
-    </x:si>
-    <x:si>
-        <x:r>
-            <x:t>String 3</x:t>
-        </x:r>
-    </x:si>
-</x:sst>"#;
-
-        let mut buf = [0; 1000];
-        let mut zip_writer = ZipWriter::new(std::io::Cursor::new(&mut buf[..]));
-        let options =
-            SimpleFileOptions::default().compression_method(zip::CompressionMethod::Stored);
-        zip_writer
-            .start_file("xl/sharedStrings.xml", options)
-            .unwrap();
-        zip_writer.write(shared_strings_data).unwrap();
-        let zip_size = zip_writer.finish().unwrap().position() as usize;
-
-        let zip = ZipArchive::new(std::io::Cursor::new(&buf[..zip_size])).unwrap();
-
-        let mut xlsx = Xlsx {
-            zip,
-            strings: vec![],
-            sheets: vec![],
-            tables: None,
-            formats: vec![],
-            is_1904: false,
-            metadata: Metadata::default(),
-            #[cfg(feature = "picture")]
-            pictures: None,
-            merged_regions: None,
-            options: XlsxOptions::default(),
-        };
-
-        assert!(xlsx.read_shared_strings().is_ok());
-        assert_eq!(3, xlsx.strings.len());
-        assert_eq!("String 1", &xlsx.strings[0]);
-        assert_eq!("String 2", &xlsx.strings[1]);
-        assert_eq!("String 3", &xlsx.strings[2]);
-    }
 }
