@@ -10,7 +10,7 @@ fn test_to_list_logical() -> PolarsResult<()> {
     // check if dtype is maintained all the way to formatting
     assert!(s.contains("[2021-01-01, 2021-01-02, 2021-01-03]"));
 
-    let expl = out.explode().unwrap();
+    let expl = out.explode(false).unwrap();
     assert_eq!(expl.dtype(), &DataType::Date);
     Ok(())
 }
