@@ -8,8 +8,9 @@ use arrow::array::*;
 use arrow::bitmap::Bitmap;
 use arrow::datatypes::*;
 use arrow::record_batch::RecordBatchT;
-use arrow::types::{i256, NativeType};
+use arrow::types::{NativeType, i256};
 use ethnum::AsI256;
+use polars::prelude::{PlSmallStr, get_column_write_options};
 use polars_error::PolarsResult;
 use polars_parquet::read::{self as p_read};
 use polars_parquet::write::*;
@@ -539,7 +540,7 @@ pub fn pyarrow_nullable(column: &str) -> Box<dyn Array> {
                 .to(ArrowDataType::Timestamp(TimeUnit::Second, None)),
         ),
         "timestamp_s_utc" => Box::new(PrimitiveArray::<i64>::from(i64_values).to(
-            ArrowDataType::Timestamp(TimeUnit::Second, Some("UTC".into())),
+            ArrowDataType::Timestamp(TimeUnit::Second, Some(PlSmallStr::from_static("UTC"))),
         )),
         _ => unreachable!(),
     }
@@ -674,30 +675,23 @@ fn integration_write(
         data_page_size: None,
     };
 
-    let encodings = schema
-        .iter_values()
-        .map(|f| {
-            transverse(&f.dtype, |x| {
-                if let ArrowDataType::Dictionary(..) = x {
-                    Encoding::RleDictionary
-                } else {
-                    Encoding::Plain
-                }
-            })
-        })
-        .collect();
+    let column_options = get_column_write_options(schema, &[]);
 
-    let row_groups =
-        RowGroupIterator::try_new(chunks.iter().cloned().map(Ok), schema, options, encodings)?;
+    let row_groups = RowGroupIterator::try_new(
+        chunks.iter().cloned().map(Ok),
+        schema,
+        options,
+        column_options.clone(),
+    )?;
 
     let writer = Cursor::new(vec![]);
 
-    let mut writer = FileWriter::try_new(writer, schema.clone(), options)?;
+    let mut writer = FileWriter::try_new(writer, schema.clone(), options, &column_options)?;
 
     for group in row_groups {
         writer.write(group?)?;
     }
-    writer.end(None)?;
+    writer.end(None, &column_options)?;
 
     Ok(writer.into_inner().into_inner())
 }
