@@ -1,9 +1,10 @@
+#![allow(unsafe_op_in_unsafe_fn)]
 #[cfg(any(feature = "fmt", feature = "fmt_no_tty"))]
 use std::borrow::Cow;
 use std::fmt::{Debug, Display, Formatter, Write};
 use std::str::FromStr;
-use std::sync::atomic::{AtomicU8, Ordering};
 use std::sync::RwLock;
+use std::sync::atomic::{AtomicU8, Ordering};
 use std::{fmt, str};
 
 #[cfg(any(
@@ -104,11 +105,7 @@ fn parse_env_var_limit(name: &str, default: usize) -> usize {
     parse_env_var(name).map_or(
         default,
         |n: i64| {
-            if n < 0 {
-                usize::MAX
-            } else {
-                n as usize
-            }
+            if n < 0 { usize::MAX } else { n as usize }
         },
     )
 }
@@ -226,7 +223,7 @@ fn format_object_array(
     array_type: &str,
 ) -> fmt::Result {
     match object.dtype() {
-        DataType::Object(inner_type, _) => {
+        DataType::Object(inner_type) => {
             let limit = std::cmp::min(DEFAULT_ROW_LIMIT, object.len());
             write!(
                 f,
@@ -251,7 +248,7 @@ where
     T: PolarsNumericType,
 {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
-        let dt = format!("{}", T::get_dtype());
+        let dt = format!("{}", T::get_static_dtype());
         format_array!(f, self, dt, self.name(), "ChunkedArray")
     }
 }
@@ -403,7 +400,7 @@ impl Debug for Series {
                 format_array!(f, self.list().unwrap(), &dt, self.name(), "Series")
             },
             #[cfg(feature = "object")]
-            DataType::Object(_, _) => format_object_array(f, self, self.name(), "Series"),
+            DataType::Object(_) => format_object_array(f, self, self.name(), "Series"),
             #[cfg(feature = "dtype-categorical")]
             DataType::Categorical(_, _) => {
                 format_array!(f, self.categorical().unwrap(), "cat", self.name(), "Series")
@@ -608,7 +605,7 @@ impl Display for DataFrame {
             let padding = 2; // eg: one char either side of the value
 
             let (n_first, n_last) = if self.width() > max_n_cols {
-                ((max_n_cols + 1) / 2, max_n_cols / 2)
+                (max_n_cols.div_ceil(2), max_n_cols / 2)
             } else {
                 (self.width(), 0)
             };
@@ -722,10 +719,15 @@ impl Display for DataFrame {
             let tbl_fallback_width = 100;
             let tbl_width = std::env::var("POLARS_TABLE_WIDTH")
                 .map(|s| {
-                    Some(
-                        s.parse::<u16>()
-                            .expect("could not parse table width argument"),
-                    )
+                    let n = s
+                        .parse::<i64>()
+                        .expect("could not parse table width argument");
+                    let w = if n < 0 {
+                        u16::MAX
+                    } else {
+                        u16::try_from(n).expect("table width argument does not fit in u16")
+                    };
+                    Some(w)
                 })
                 .unwrap_or(None);
 
@@ -742,8 +744,12 @@ impl Display for DataFrame {
                     str_truncate + ellipsis_len + padding,
                     std::cmp::max(name_lengths[idx], *elem_len),
                 );
-                if mx <= min_col_width {
+                if (mx <= min_col_width) && !(max_n_rows > 0 && height > max_n_rows) {
+                    // col width is less than min width + table is not truncated
                     constraints.push(col_width_exact(mx));
+                } else if mx <= min_col_width {
+                    // col width is less than min width + table is truncated (w/ ellipsis)
+                    constraints.push(col_width_bounds(mx, min_col_width));
                 } else {
                     constraints.push(col_width_bounds(min_col_width, mx));
                 }
@@ -830,15 +836,14 @@ fn fmt_int_string_custom(num: &str, group_size: u8, group_separator: &str) -> St
         } else {
             0
         };
-        let int_body = num[sign_offset..]
-            .as_bytes()
+        let int_body = &num.as_bytes()[sign_offset..]
             .rchunks(group_size as usize)
             .rev()
             .map(str::from_utf8)
             .collect::<Result<Vec<&str>, _>>()
             .unwrap()
             .join(group_separator);
-        out.push_str(&int_body);
+        out.push_str(int_body);
         out
     }
 }
@@ -1266,7 +1271,7 @@ impl Series {
             _ if max_items >= self.len() => {
                 // this will always leave a trailing ", " after the last item
                 // but for long lists, this is faster than checking against the length each time
-                for item in self.iter() {
+                for item in self.rechunk().iter() {
                     write!(result, "{item}, ").unwrap();
                 }
                 // remove trailing ", " and replace with closing brace
@@ -1304,6 +1309,7 @@ fn fmt_decimal(f: &mut Formatter<'_>, v: i128, scale: usize) -> fmt::Result {
     feature = "dtype-date",
     feature = "dtype-datetime"
 ))]
+#[allow(unsafe_op_in_unsafe_fn)]
 mod test {
     use crate::prelude::*;
 
@@ -1329,7 +1335,7 @@ Series: 'a' [list[i32]]
             format!("{:?}", list_long)
         );
 
-        std::env::set_var("POLARS_FMT_TABLE_CELL_LIST_LEN", "10");
+        unsafe { std::env::set_var("POLARS_FMT_TABLE_CELL_LIST_LEN", "10") };
 
         assert_eq!(
             r#"shape: (2,)
@@ -1341,7 +1347,7 @@ Series: 'a' [list[i32]]
             format!("{:?}", list_long)
         );
 
-        std::env::set_var("POLARS_FMT_TABLE_CELL_LIST_LEN", "-1");
+        unsafe { std::env::set_var("POLARS_FMT_TABLE_CELL_LIST_LEN", "-1") };
 
         assert_eq!(
             r#"shape: (2,)
@@ -1353,7 +1359,7 @@ Series: 'a' [list[i32]]
             format!("{:?}", list_long)
         );
 
-        std::env::set_var("POLARS_FMT_TABLE_CELL_LIST_LEN", "0");
+        unsafe { std::env::set_var("POLARS_FMT_TABLE_CELL_LIST_LEN", "0") };
 
         assert_eq!(
             r#"shape: (2,)
@@ -1365,7 +1371,7 @@ Series: 'a' [list[i32]]
             format!("{:?}", list_long)
         );
 
-        std::env::set_var("POLARS_FMT_TABLE_CELL_LIST_LEN", "1");
+        unsafe { std::env::set_var("POLARS_FMT_TABLE_CELL_LIST_LEN", "1") };
 
         assert_eq!(
             r#"shape: (2,)
@@ -1377,7 +1383,7 @@ Series: 'a' [list[i32]]
             format!("{:?}", list_long)
         );
 
-        std::env::set_var("POLARS_FMT_TABLE_CELL_LIST_LEN", "4");
+        unsafe { std::env::set_var("POLARS_FMT_TABLE_CELL_LIST_LEN", "4") };
 
         assert_eq!(
             r#"shape: (2,)
@@ -1399,7 +1405,7 @@ Series: 'a' [list[i32]]
         builder.append_opt_slice(None);
         let list_short = builder.finish().into_series();
 
-        std::env::set_var("POLARS_FMT_TABLE_CELL_LIST_LEN", "");
+        unsafe { std::env::set_var("POLARS_FMT_TABLE_CELL_LIST_LEN", "") };
 
         assert_eq!(
             r#"shape: (2,)
@@ -1411,7 +1417,7 @@ Series: 'a' [list[i32]]
             format!("{:?}", list_short)
         );
 
-        std::env::set_var("POLARS_FMT_TABLE_CELL_LIST_LEN", "0");
+        unsafe { std::env::set_var("POLARS_FMT_TABLE_CELL_LIST_LEN", "0") };
 
         assert_eq!(
             r#"shape: (2,)
@@ -1423,7 +1429,7 @@ Series: 'a' [list[i32]]
             format!("{:?}", list_short)
         );
 
-        std::env::set_var("POLARS_FMT_TABLE_CELL_LIST_LEN", "-1");
+        unsafe { std::env::set_var("POLARS_FMT_TABLE_CELL_LIST_LEN", "-1") };
 
         assert_eq!(
             r#"shape: (2,)
@@ -1445,7 +1451,7 @@ Series: 'a' [list[i32]]
         builder.append_opt_slice(None);
         let list_empty = builder.finish().into_series();
 
-        std::env::set_var("POLARS_FMT_TABLE_CELL_LIST_LEN", "");
+        unsafe { std::env::set_var("POLARS_FMT_TABLE_CELL_LIST_LEN", "") };
 
         assert_eq!(
             r#"shape: (2,)
