@@ -164,7 +164,7 @@ pub mod checked {
             lhs: &ChunkedArray<Self>,
             _rhs: T,
         ) -> PolarsResult<Series> {
-            polars_bail!(opq = checked_div_num, lhs.dtype(), Self::get_dtype());
+            polars_bail!(opq = checked_div_num, lhs.dtype(), Self::get_static_dtype());
         }
     }
 
@@ -310,11 +310,7 @@ pub mod checked {
                     .apply(|opt_v| {
                         opt_v.and_then(|v| {
                             let res = rhs.to_f32().unwrap();
-                            if res.is_zero() {
-                                None
-                            } else {
-                                Some(v / res)
-                            }
+                            if res.is_zero() { None } else { Some(v / res) }
                         })
                     })
                     .into_series(),
@@ -324,11 +320,7 @@ pub mod checked {
                     .apply(|opt_v| {
                         opt_v.and_then(|v| {
                             let res = rhs.to_f64().unwrap();
-                            if res.is_zero() {
-                                None
-                            } else {
-                                Some(v / res)
-                            }
+                            if res.is_zero() { None } else { Some(v / res) }
                         })
                     })
                     .into_series(),
@@ -448,8 +440,21 @@ pub fn _struct_arithmetic<F: FnMut(&Series, &Series) -> PolarsResult<Series>>(
             Ok(rhs.try_apply_fields(|rhs| func(s, rhs))?.into_series())
         },
         _ => {
-            let (s, rhs) = align_chunks_binary(s, rhs);
+            let mut s = Cow::Borrowed(s);
+            let mut rhs = Cow::Borrowed(rhs);
+
+            match (s.len(), rhs.len()) {
+                (l, r) if l == r => {},
+                (1, _) => s = Cow::Owned(s.new_from_index(0, rhs.len())),
+                (_, 1) => rhs = Cow::Owned(rhs.new_from_index(0, s.len())),
+                (l, r) => {
+                    polars_bail!(ComputeError: "Struct arithmetic between different lengths {l} != {r}")
+                },
+            };
+            let (s, rhs) = align_chunks_binary(&s, &rhs);
             let mut s = s.into_owned();
+
+            // Expects lengths to be equal.
             s.zip_outer_validity(rhs.as_ref());
 
             let mut rhs_iter = rhs.fields_as_series().into_iter();
@@ -659,9 +664,7 @@ where
     fn sub(self, rhs: T) -> Self::Output {
         let s = self.to_physical_repr();
         macro_rules! sub {
-            ($ca:expr) => {{
-                $ca.sub(rhs).into_series()
-            }};
+            ($ca:expr) => {{ $ca.sub(rhs).into_series() }};
         }
 
         let out = downcast_as_macro_arg_physical!(s, sub);
@@ -689,9 +692,7 @@ where
     fn add(self, rhs: T) -> Self::Output {
         let s = self.to_physical_repr();
         macro_rules! add {
-            ($ca:expr) => {{
-                $ca.add(rhs).into_series()
-            }};
+            ($ca:expr) => {{ $ca.add(rhs).into_series() }};
         }
         let out = downcast_as_macro_arg_physical!(s, add);
         finish_cast(self, out)
@@ -718,9 +719,7 @@ where
     fn div(self, rhs: T) -> Self::Output {
         let s = self.to_physical_repr();
         macro_rules! div {
-            ($ca:expr) => {{
-                $ca.div(rhs).into_series()
-            }};
+            ($ca:expr) => {{ $ca.div(rhs).into_series() }};
         }
 
         let out = downcast_as_macro_arg_physical!(s, div);
@@ -764,9 +763,7 @@ where
     fn mul(self, rhs: T) -> Self::Output {
         let s = self.to_physical_repr();
         macro_rules! mul {
-            ($ca:expr) => {{
-                $ca.mul(rhs).into_series()
-            }};
+            ($ca:expr) => {{ $ca.mul(rhs).into_series() }};
         }
         let out = downcast_as_macro_arg_physical!(s, mul);
         finish_cast(self, out)
@@ -793,9 +790,7 @@ where
     fn rem(self, rhs: T) -> Self::Output {
         let s = self.to_physical_repr();
         macro_rules! rem {
-            ($ca:expr) => {{
-                $ca.rem(rhs).into_series()
-            }};
+            ($ca:expr) => {{ $ca.rem(rhs).into_series() }};
         }
         let out = downcast_as_macro_arg_physical!(s, rem);
         finish_cast(self, out)
@@ -866,9 +861,7 @@ where
     fn sub(self, rhs: &Series) -> Self::Output {
         let s = rhs.to_physical_repr();
         macro_rules! sub {
-            ($rhs:expr) => {{
-                $rhs.lhs_sub(self).into_series()
-            }};
+            ($rhs:expr) => {{ $rhs.lhs_sub(self).into_series() }};
         }
         let out = downcast_as_macro_arg_physical!(s, sub);
 
@@ -877,9 +870,7 @@ where
     fn div(self, rhs: &Series) -> Self::Output {
         let s = rhs.to_physical_repr();
         macro_rules! div {
-            ($rhs:expr) => {{
-                $rhs.lhs_div(self).into_series()
-            }};
+            ($rhs:expr) => {{ $rhs.lhs_div(self).into_series() }};
         }
         let out = downcast_as_macro_arg_physical!(s, div);
 
@@ -892,9 +883,7 @@ where
     fn rem(self, rhs: &Series) -> Self::Output {
         let s = rhs.to_physical_repr();
         macro_rules! rem {
-            ($rhs:expr) => {{
-                $rhs.lhs_rem(self).into_series()
-            }};
+            ($rhs:expr) => {{ $rhs.lhs_rem(self).into_series() }};
         }
 
         let out = downcast_as_macro_arg_physical!(s, rem);
