@@ -2,6 +2,7 @@ use std::any::Any;
 use std::borrow::Cow;
 
 use arrow::bitmap::{Bitmap, BitmapBuilder};
+use polars_compute::rolling::QuantileMethod;
 #[cfg(feature = "serde")]
 use serde::{Deserialize, Serialize};
 
@@ -35,6 +36,8 @@ pub enum BitRepr {
 }
 
 pub(crate) mod private {
+    use polars_utils::aliases::PlSeedableRandomStateQuality;
+
     use super::*;
     use crate::chunked_array::flags::StatisticsFlags;
     use crate::chunked_array::ops::compare_inner::{TotalEqInner, TotalOrdInner};
@@ -81,12 +84,16 @@ pub(crate) mod private {
         #[expect(clippy::wrong_self_convention)]
         fn into_total_ord_inner<'a>(&'a self) -> Box<dyn TotalOrdInner + 'a>;
 
-        fn vec_hash(&self, _build_hasher: PlRandomState, _buf: &mut Vec<u64>) -> PolarsResult<()> {
+        fn vec_hash(
+            &self,
+            _build_hasher: PlSeedableRandomStateQuality,
+            _buf: &mut Vec<u64>,
+        ) -> PolarsResult<()> {
             polars_bail!(opq = vec_hash, self._dtype());
         }
         fn vec_hash_combine(
             &self,
-            _build_hasher: PlRandomState,
+            _build_hasher: PlSeedableRandomStateQuality,
             _hashes: &mut [u64],
         ) -> PolarsResult<()> {
             polars_bail!(opq = vec_hash_combine, self._dtype());
@@ -371,6 +378,25 @@ pub trait SeriesTrait:
     /// ```
     fn new_from_index(&self, _index: usize, _length: usize) -> Series;
 
+    /// Trim all lists of unused start and end elements recursively.
+    ///
+    /// - `None` if nothing needed to be done.
+    /// - `Some(series)` if something changed.
+    fn trim_lists_to_normalized_offsets(&self) -> Option<Series> {
+        None
+    }
+
+    /// Propagate down nulls in nested types.
+    ///
+    /// - `None` if nothing needed to be done.
+    /// - `Some(series)` if something changed.
+    fn propagate_nulls(&self) -> Option<Series> {
+        None
+    }
+
+    /// Find the indices of elements where the null masks are different recursively.
+    fn find_validity_mismatch(&self, other: &Series, idxs: &mut Vec<IdxSize>);
+
     fn cast(&self, _dtype: &DataType, options: CastOptions) -> PolarsResult<Series>;
 
     /// Get a single value by index. Don't use this operation for loops as a runtime cast is
@@ -571,6 +597,10 @@ pub trait SeriesTrait:
     /// reference.
     fn as_any_mut(&mut self) -> &mut dyn Any;
 
+    /// Get a hold of the [`ChunkedArray`] or `NullChunked` as an `Any` trait reference. This
+    /// pierces through `Logical` types to get the underlying physical array.
+    fn as_phys_any(&self) -> &dyn Any;
+
     fn as_arc_any(self: Arc<Self>) -> Arc<dyn Any + Send + Sync>;
 
     #[cfg(feature = "checked_arithmetic")]
@@ -595,7 +625,7 @@ impl (dyn SeriesTrait + '_) {
     where
         N: 'static + PolarsDataType<IsLogical = FalseT>,
     {
-        polars_ensure!(&N::get_dtype() == self.dtype(), unpack);
+        polars_ensure!(&N::get_static_dtype() == self.dtype(), unpack);
         Ok(self.as_ref())
     }
 }
