@@ -4,16 +4,17 @@ use arrow::array::*;
 use arrow::compute::concatenate::concatenate;
 use arrow::legacy::utils::CustomIterTools;
 use arrow::offset::Offsets;
+use polars_compute::rolling::QuantileMethod;
+use polars_core::POOL;
 use polars_core::prelude::*;
 use polars_core::series::IsSorted;
-use polars_core::utils::{NoNull, _split_offsets};
-use polars_core::POOL;
+use polars_core::utils::{_split_offsets, NoNull};
 #[cfg(feature = "propagate_nans")]
 use polars_ops::prelude::nan_propagating_aggregate;
 use rayon::prelude::*;
 
 use super::*;
-use crate::expressions::AggState::{AggregatedList, AggregatedScalar};
+use crate::expressions::AggState::AggregatedScalar;
 use crate::expressions::{
     AggState, AggregationContext, PartitionedAggregation, PhysicalExpr, UpdateGroups,
 };
@@ -163,7 +164,9 @@ impl PhysicalExpr for AggregationExpr {
         let mut ac = self.input.evaluate_on_groups(df, groups, state)?;
         // don't change names by aggregations as is done in polars-core
         let keep_name = ac.get_values().name().clone();
-        polars_ensure!(!matches!(ac.agg_state(), AggState::Literal(_)), ComputeError: "cannot aggregate a literal");
+
+        // Literals cannot be aggregated except for implode.
+        polars_ensure!((!matches!(ac.agg_state(), AggState::Literal(_)) || matches!(self.agg_type.groupby, GroupByMethod::Implode)), ComputeError: "cannot aggregate a literal");
 
         if let AggregatedScalar(_) = ac.agg_state() {
             match self.agg_type.groupby {
@@ -353,18 +356,23 @@ impl PhysicalExpr for AggregationExpr {
                     let c = match ac.agg_state() {
                         // mean agg:
                         // -> f64 -> list<f64>
-                        AggState::AggregatedScalar(c) => c
+                        AggregatedScalar(c) => c
                             .reshape_list(&[
                                 ReshapeDimension::Infer,
                                 ReshapeDimension::new_dimension(1),
                             ])
                             .unwrap(),
+                        // Auto-imploded
+                        AggState::NotAggregated(_) | AggState::AggregatedList(_) => {
+                            ac._implode_no_agg();
+                            return Ok(ac);
+                        },
                         _ => {
                             let agg = ac.aggregated();
                             agg.as_list().into_column()
                         },
                     };
-                    AggregatedList(c.with_name(keep_name))
+                    AggState::AggregatedList(c.with_name(keep_name))
                 },
                 GroupByMethod::Groups => {
                     let mut column: ListChunked = ac.groups().as_list_chunked();
@@ -442,10 +450,6 @@ impl PhysicalExpr for AggregationExpr {
         }
     }
 
-    fn collect_live_columns(&self, lv: &mut PlIndexSet<PlSmallStr>) {
-        self.input.collect_live_columns(lv);
-    }
-
     fn is_scalar(&self) -> bool {
         true
     }
@@ -570,7 +574,13 @@ impl PartitionedAggregation for AggregationExpr {
                         let count = &fields[1];
                         let (agg_count, agg_s) =
                             unsafe { POOL.join(|| count.agg_sum(groups), || sum.agg_sum(groups)) };
-                        let agg_s = &agg_s / &agg_count;
+
+                        // Ensure that we don't divide by zero by masking out zeros.
+                        let agg_count = agg_count.idx().unwrap();
+                        let mask = agg_count.equal(0 as IdxSize);
+                        let agg_count = agg_count.set(&mask, None).unwrap().into_series();
+
+                        let agg_s = &agg_s / &agg_count.cast(agg_s.dtype()).unwrap();
                         Ok(agg_s?.with_name(new_name).into_column())
                     },
                     _ => Ok(Column::full_null(
@@ -594,12 +604,12 @@ impl PartitionedAggregation for AggregationExpr {
                 offsets.push(length_so_far);
 
                 let mut process_group = |ca: ListChunked| -> PolarsResult<()> {
-                    let s = ca.explode()?;
+                    let s = ca.explode(false)?;
                     length_so_far += s.len() as i64;
                     offsets.push(length_so_far);
                     values.push(s.chunks()[0].clone());
 
-                    if s.len() == 0 {
+                    if s.is_empty() {
                         can_fast_explode = false;
                     }
                     Ok(())
@@ -735,11 +745,6 @@ impl PhysicalExpr for AggQuantileExpr {
         ))
     }
 
-    fn collect_live_columns(&self, lv: &mut PlIndexSet<PlSmallStr>) {
-        self.input.collect_live_columns(lv);
-        self.quantile.collect_live_columns(lv);
-    }
-
     fn to_field(&self, input_schema: &Schema) -> PolarsResult<Field> {
         self.input.to_field(input_schema)
     }
@@ -763,10 +768,17 @@ where
     #[cfg(not(debug_assertions))]
     let thread_boundary = 100_000;
 
+    // Temporary until categorical min/max multithreading implementation is corrected.
+    #[cfg(feature = "dtype-categorical")]
+    let is_categorical = matches!(s.dtype(), &DataType::Categorical(_, _));
+    #[cfg(not(feature = "dtype-categorical"))]
+    let is_categorical = false;
     // threading overhead/ splitting work stealing is costly..
-    if allow_threading
+
+    if !allow_threading
         || s.len() < thread_boundary
         || POOL.current_thread_has_pending_tasks().unwrap_or(false)
+        || is_categorical
     {
         return f(s);
     }
