@@ -5,9 +5,9 @@ use arrow::bitmap::Bitmap;
 use polars_core::prelude::*;
 use polars_core::series::IsSorted;
 use polars_core::utils::_split_offsets;
-use polars_core::{downcast_as_macro_arg_physical, POOL};
-use polars_ops::frame::join::{private_left_join_multiple_keys, ChunkJoinOptIds};
+use polars_core::{POOL, downcast_as_macro_arg_physical};
 use polars_ops::frame::SeriesJoin;
+use polars_ops::frame::join::{ChunkJoinOptIds, private_left_join_multiple_keys};
 use polars_ops::prelude::*;
 use polars_plan::prelude::*;
 use polars_utils::sort::perfect_sort;
@@ -306,9 +306,7 @@ impl WindowExpr {
                         },
                         Expr::Function { options, .. }
                         | Expr::AnonymousFunction { options, .. } => {
-                            if options.flags.contains(FunctionFlags::RETURNS_SCALAR)
-                                && matches!(options.collect_groups, ApplyOptions::GroupWise)
-                            {
+                            if options.flags.returns_scalar() {
                                 agg_col = true;
                             }
                         },
@@ -408,7 +406,18 @@ impl PhysicalExpr for WindowExpr {
 
         if df.is_empty() {
             let field = self.phys_function.to_field(df.schema())?;
-            return Ok(Column::full_null(field.name().clone(), 0, field.dtype()));
+            match self.mapping {
+                WindowMapping::Join => {
+                    return Ok(Column::full_null(
+                        field.name().clone(),
+                        0,
+                        &DataType::List(Box::new(field.dtype().clone())),
+                    ));
+                },
+                _ => {
+                    return Ok(Column::full_null(field.name().clone(), 0, field.dtype()));
+                },
+            }
         }
 
         let group_by_columns = self
@@ -522,7 +531,7 @@ impl PhysicalExpr for WindowExpr {
                 Ok(out.into_column())
             },
             Explode => {
-                let mut out = ac.aggregated().explode()?;
+                let mut out = ac.aggregated().explode(false)?;
                 if let Some(name) = &self.out_name {
                     out.rename(name.clone());
                 }
@@ -532,7 +541,7 @@ impl PhysicalExpr for WindowExpr {
                 // TODO!
                 // investigate if sorted arrays can be return directly
                 let out_column = ac.aggregated();
-                let flattened = out_column.explode()?;
+                let flattened = out_column.explode(false)?;
                 // we extend the lifetime as we must convince the compiler that ac lives
                 // long enough. We drop `GrouBy` when we are done with `ac`.
                 let ac = unsafe {
@@ -557,7 +566,7 @@ impl PhysicalExpr for WindowExpr {
                 let update_groups = !matches!(&ac.update_groups, UpdateGroups::No);
                 match (
                     &ac.update_groups,
-                    set_by_groups(&out_column, &ac.groups, df.height(), update_groups),
+                    set_by_groups(&out_column, &ac, df.height(), update_groups),
                 ) {
                     // for aggregations that reduce like sum, mean, first and are numeric
                     // we take the group locations to directly map them to the right place
@@ -650,16 +659,6 @@ impl PhysicalExpr for WindowExpr {
         false
     }
 
-    fn collect_live_columns(&self, lv: &mut PlIndexSet<PlSmallStr>) {
-        for i in &self.group_by {
-            i.collect_live_columns(lv);
-        }
-        if let Some((i, _)) = &self.order_by {
-            i.collect_live_columns(lv);
-        }
-        self.phys_function.collect_live_columns(lv);
-    }
-
     #[allow(clippy::ptr_arg)]
     fn evaluate_on_groups<'a>(
         &self,
@@ -684,7 +683,7 @@ fn materialize_column(join_opt_ids: &ChunkJoinOptIds, out_column: &Column) -> Co
             Either::Left(ids) => unsafe {
                 IdxCa::with_nullable_idx(ids, |idx| out_column.take_unchecked(idx))
             },
-            Either::Right(ids) => unsafe { out_column.take_opt_chunked_unchecked(ids) },
+            Either::Right(ids) => unsafe { out_column.take_opt_chunked_unchecked(ids, false) },
         }
     }
 }
@@ -692,11 +691,11 @@ fn materialize_column(join_opt_ids: &ChunkJoinOptIds, out_column: &Column) -> Co
 /// Simple reducing aggregation can be set by the groups
 fn set_by_groups(
     s: &Column,
-    groups: &GroupsType,
+    ac: &AggregationContext,
     len: usize,
     update_groups: bool,
 ) -> Option<Column> {
-    if update_groups {
+    if update_groups || !ac.original_len {
         return None;
     }
     if s.dtype().to_physical().is_primitive_numeric() {
@@ -704,12 +703,10 @@ fn set_by_groups(
         let s = s.to_physical_repr();
 
         macro_rules! dispatch {
-            ($ca:expr) => {{
-                Some(set_numeric($ca, groups, len))
-            }};
+            ($ca:expr) => {{ Some(set_numeric($ca, &ac.groups, len)) }};
         }
         downcast_as_macro_arg_physical!(&s, dispatch)
-            .map(|s| s.cast(dtype).unwrap())
+            .map(|s| unsafe { s.from_physical_unchecked(dtype) }.unwrap())
             .map(Column::from)
     } else {
         None
@@ -840,7 +837,9 @@ where
         unsafe { values.set_len(len) }
         let validity = Bitmap::from(validity);
         let arr = PrimitiveArray::new(
-            T::get_dtype().to_physical().to_arrow(CompatLevel::newest()),
+            T::get_static_dtype()
+                .to_physical()
+                .to_arrow(CompatLevel::newest()),
             values.into(),
             Some(validity),
         );
