@@ -3,16 +3,17 @@ use std::path::Path;
 use memchr::memchr2_iter;
 use num_traits::Pow;
 use polars_core::prelude::*;
-use polars_core::{config, POOL};
+use polars_core::{POOL, config};
 use polars_error::feature_gated;
-use polars_utils::index::Bounded;
+use polars_utils::mmap::MMapSemaphore;
+use polars_utils::select::select_unpredictable;
 use rayon::prelude::*;
 
+use super::CsvParseOptions;
 use super::buffer::Buffer;
 use super::options::{CommentPrefix, NullValuesCompiled};
 use super::splitfields::SplitFields;
 use super::utils::get_file_chunks;
-use super::CsvParseOptions;
 use crate::path_utils::is_cloud_url;
 use crate::utils::compression::maybe_decompress_bytes;
 
@@ -38,11 +39,11 @@ pub fn count_rows(
         polars_utils::open_file(path)?
     };
 
-    let mmap = unsafe { memmap::Mmap::map(&file).unwrap() };
+    let mmap = MMapSemaphore::new_from_file(&file).unwrap();
     let owned = &mut vec![];
     let reader_bytes = maybe_decompress_bytes(mmap.as_ref(), owned)?;
 
-    count_rows_from_slice(
+    count_rows_from_slice_par(
         reader_bytes,
         separator,
         quote_char,
@@ -54,7 +55,7 @@ pub fn count_rows(
 
 /// Read the number of rows without parsing columns
 /// useful for count(*) queries
-pub fn count_rows_from_slice(
+pub fn count_rows_from_slice_par(
     mut bytes: &[u8],
     separator: u8,
     quote_char: Option<u8>,
@@ -88,27 +89,57 @@ pub fn count_rows_from_slice(
     })
     .unwrap_or(1);
 
+    if n_threads == 1 {
+        return count_rows_from_slice(bytes, quote_char, comment_prefix, eol_char, has_header);
+    }
+
     let file_chunks: Vec<(usize, usize)> =
         get_file_chunks(bytes, n_threads, None, separator, quote_char, eol_char);
 
     let iter = file_chunks.into_par_iter().map(|(start, stop)| {
-        let local_bytes = &bytes[start..stop];
-        let row_iterator = SplitLines::new(local_bytes, quote_char, eol_char, comment_prefix);
+        let bytes = &bytes[start..stop];
+
         if comment_prefix.is_some() {
-            Ok(row_iterator
-                .filter(|line| !line.is_empty() && !is_comment_line(line, comment_prefix))
-                .count())
+            SplitLines::new(bytes, quote_char, eol_char, comment_prefix)
+                .filter(|line| !is_comment_line(line, comment_prefix))
+                .count()
         } else {
-            Ok(row_iterator.count())
+            CountLines::new(quote_char, eol_char).count(bytes).0
+                + bytes.last().is_some_and(|x| *x != b'\n') as usize
         }
     });
 
-    let count_result: PolarsResult<usize> = POOL.install(|| iter.sum());
+    let n: usize = POOL.install(|| iter.sum());
 
-    match count_result {
-        Ok(val) => Ok(val - (has_header as usize)),
-        Err(err) => Err(err),
+    Ok(n - (has_header as usize))
+}
+
+/// Read the number of rows without parsing columns
+pub fn count_rows_from_slice(
+    mut bytes: &[u8],
+    quote_char: Option<u8>,
+    comment_prefix: Option<&CommentPrefix>,
+    eol_char: u8,
+    has_header: bool,
+) -> PolarsResult<usize> {
+    for _ in 0..bytes.len() {
+        if bytes[0] != eol_char {
+            break;
+        }
+
+        bytes = &bytes[1..];
     }
+
+    let n = if comment_prefix.is_some() {
+        SplitLines::new(bytes, quote_char, eol_char, comment_prefix)
+            .filter(|line| !is_comment_line(line, comment_prefix))
+            .count()
+    } else {
+        CountLines::new(quote_char, eol_char).count(bytes).0
+            + bytes.last().is_some_and(|x| *x != b'\n') as usize
+    };
+
+    Ok(n - (has_header as usize))
 }
 
 /// Skip the utf-8 Byte Order Mark.
@@ -607,6 +638,13 @@ pub struct CountLines {
     quoting: bool,
 }
 
+#[derive(Copy, Clone, Debug)]
+pub struct LineStats {
+    newline_count: usize,
+    last_newline_offset: usize,
+    end_inside_string: bool,
+}
+
 impl CountLines {
     pub fn new(quote_char: Option<u8>, eol_char: u8) -> Self {
         let quoting = quote_char.is_some();
@@ -626,6 +664,98 @@ impl CountLines {
         }
     }
 
+    /// Analyzes a chunk of CSV data.
+    ///
+    /// Returns (newline_count, last_newline_offset, end_inside_string) twice,
+    /// the first is assuming the start of the chunk is *not* inside a string,
+    /// the second assuming the start is inside a string.
+    pub fn analyze_chunk(&self, bytes: &[u8]) -> [LineStats; 2] {
+        let mut scan_offset = 0;
+        let mut states = [
+            LineStats {
+                newline_count: 0,
+                last_newline_offset: 0,
+                end_inside_string: false,
+            },
+            LineStats {
+                newline_count: 0,
+                last_newline_offset: 0,
+                end_inside_string: false,
+            },
+        ];
+
+        // false if even number of quotes seen so far, true otherwise.
+        #[allow(unused_assignments)]
+        let mut global_quote_parity = false;
+
+        #[cfg(feature = "simd")]
+        {
+            // 0 if even number of quotes seen so far, u64::MAX otherwise.
+            let mut global_quote_parity_mask = 0;
+            while scan_offset + 64 <= bytes.len() {
+                let block: [u8; 64] = unsafe {
+                    bytes
+                        .get_unchecked(scan_offset..scan_offset + 64)
+                        .try_into()
+                        .unwrap_unchecked()
+                };
+                let simd_bytes = SimdVec::from(block);
+                let eol_mask = simd_bytes.simd_eq(self.simd_eol_char).to_bitmask();
+                if self.quoting {
+                    let quote_mask = simd_bytes.simd_eq(self.simd_quote_char).to_bitmask();
+                    let quote_parity =
+                        prefix_xorsum_inclusive(quote_mask) ^ global_quote_parity_mask;
+                    global_quote_parity_mask = ((quote_parity as i64) >> 63) as u64;
+
+                    let start_outside_string_eol_mask = eol_mask & !quote_parity;
+                    states[0].newline_count += start_outside_string_eol_mask.count_ones() as usize;
+                    states[0].last_newline_offset = select_unpredictable(
+                        start_outside_string_eol_mask != 0,
+                        (scan_offset + 63)
+                            .wrapping_sub(start_outside_string_eol_mask.leading_zeros() as usize),
+                        states[0].last_newline_offset,
+                    );
+
+                    let start_inside_string_eol_mask = eol_mask & quote_parity;
+                    states[1].newline_count += start_inside_string_eol_mask.count_ones() as usize;
+                    states[1].last_newline_offset = select_unpredictable(
+                        start_inside_string_eol_mask != 0,
+                        (scan_offset + 63)
+                            .wrapping_sub(start_inside_string_eol_mask.leading_zeros() as usize),
+                        states[1].last_newline_offset,
+                    );
+                } else {
+                    states[0].newline_count += eol_mask.count_ones() as usize;
+                    states[0].last_newline_offset = select_unpredictable(
+                        eol_mask != 0,
+                        (scan_offset + 63).wrapping_sub(eol_mask.leading_zeros() as usize),
+                        states[0].last_newline_offset,
+                    );
+                }
+
+                scan_offset += 64;
+            }
+
+            global_quote_parity = global_quote_parity_mask > 0;
+        }
+
+        while scan_offset < bytes.len() {
+            let c = unsafe { *bytes.get_unchecked(scan_offset) };
+            global_quote_parity ^= (c == self.quote_char) & self.quoting;
+
+            let state = &mut states[global_quote_parity as usize];
+            state.newline_count += (c == self.eol_char) as usize;
+            state.last_newline_offset =
+                select_unpredictable(c == self.eol_char, scan_offset, state.last_newline_offset);
+
+            scan_offset += 1;
+        }
+
+        states[0].end_inside_string = global_quote_parity;
+        states[1].end_inside_string = !global_quote_parity;
+        states
+    }
+
     pub fn find_next(&self, bytes: &[u8], chunk_size: &mut usize) -> (usize, usize) {
         loop {
             let b = unsafe { bytes.get_unchecked(..(*chunk_size).min(bytes.len())) };
@@ -640,7 +770,7 @@ impl CountLines {
         }
     }
 
-    // Returns count and offset in slice
+    /// Returns count and offset to split for remainder in slice.
     #[cfg(feature = "simd")]
     pub fn count(&self, bytes: &[u8]) -> (usize, usize) {
         let mut total_idx = 0;
@@ -779,10 +909,10 @@ pub(super) fn skip_this_line_naive(input: &[u8], eol_char: u8) -> &[u8] {
 /// # Arguments
 /// * `bytes` - input to parse
 /// * `offset` - offset in bytes in total input. This is 0 if single threaded. If multi-threaded every
-///              thread has a different offset.
+///   thread has a different offset.
 /// * `projection` - Indices of the columns to project.
 /// * `buffers` - Parsed output will be written to these buffers. Except for UTF8 data. The offsets of the
-///               fields are written to the buffers. The UTF8 data will be parsed later.
+///   fields are written to the buffers. The UTF8 data will be parsed later.
 #[allow(clippy::too_many_arguments)]
 pub(super) fn parse_lines(
     mut bytes: &[u8],
