@@ -1,3 +1,4 @@
+use std::borrow::Cow;
 use std::fmt::Write;
 
 use arrow::array::ValueSize;
@@ -300,6 +301,22 @@ pub trait ListNameSpaceImpl: AsList {
         let ca = self.as_list();
         let periods_s = periods.cast(&DataType::Int64)?;
         let periods = periods_s.i64()?;
+
+        polars_ensure!(
+            ca.len() == periods.len() || ca.len() == 1 || periods.len() == 1,
+            length_mismatch = "list.shift",
+            ca.len(),
+            periods.len()
+        );
+
+        // Broadcast `self`
+        let mut ca = Cow::Borrowed(ca);
+        if ca.len() == 1 && periods.len() != 1 {
+            // Optimize: Don't broadcast and instead have a special path.
+            ca = Cow::Owned(ca.new_from_index(0, periods.len()));
+        }
+        let ca = ca.as_ref();
+
         let out = match periods.len() {
             1 => {
                 if let Some(periods) = periods.get(0) {
@@ -372,8 +389,9 @@ pub trait ListNameSpaceImpl: AsList {
         let list_ca = self.as_list();
         let out = match (n.len(), offset.len()) {
             (1, 1) => match (n.get(0), offset.get(0)) {
-                (Some(n), Some(offset)) => list_ca
-                    .apply_amortized(|s| s.as_ref().gather_every(n as usize, offset as usize)),
+                (Some(n), Some(offset)) => list_ca.try_apply_amortized(|s| {
+                    s.as_ref().gather_every(n as usize, offset as usize)
+                })?,
                 _ => ListChunked::full_null_with_dtype(
                     list_ca.name().clone(),
                     list_ca.len(),
@@ -382,14 +400,14 @@ pub trait ListNameSpaceImpl: AsList {
             },
             (1, len_offset) if len_offset == list_ca.len() => {
                 if let Some(n) = n.get(0) {
-                    list_ca.zip_and_apply_amortized(offset, |opt_s, opt_offset| {
+                    list_ca.try_zip_and_apply_amortized(offset, |opt_s, opt_offset| {
                         match (opt_s, opt_offset) {
                             (Some(s), Some(offset)) => {
-                                Some(s.as_ref().gather_every(n as usize, offset as usize))
+                                Ok(Some(s.as_ref().gather_every(n as usize, offset as usize)?))
                             },
-                            _ => None,
+                            _ => Ok(None),
                         }
-                    })
+                    })?
                 } else {
                     ListChunked::full_null_with_dtype(
                         list_ca.name().clone(),
@@ -400,12 +418,12 @@ pub trait ListNameSpaceImpl: AsList {
             },
             (len_n, 1) if len_n == list_ca.len() => {
                 if let Some(offset) = offset.get(0) {
-                    list_ca.zip_and_apply_amortized(n, |opt_s, opt_n| match (opt_s, opt_n) {
+                    list_ca.try_zip_and_apply_amortized(n, |opt_s, opt_n| match (opt_s, opt_n) {
                         (Some(s), Some(n)) => {
-                            Some(s.as_ref().gather_every(n as usize, offset as usize))
+                            Ok(Some(s.as_ref().gather_every(n as usize, offset as usize)?))
                         },
-                        _ => None,
-                    })
+                        _ => Ok(None),
+                    })?
                 } else {
                     ListChunked::full_null_with_dtype(
                         list_ca.name().clone(),
@@ -415,14 +433,16 @@ pub trait ListNameSpaceImpl: AsList {
                 }
             },
             (len_n, len_offset) if len_n == len_offset && len_n == list_ca.len() => list_ca
-                .binary_zip_and_apply_amortized(n, offset, |opt_s, opt_n, opt_offset| {
-                    match (opt_s, opt_n, opt_offset) {
+                .try_binary_zip_and_apply_amortized(
+                    n,
+                    offset,
+                    |opt_s, opt_n, opt_offset| match (opt_s, opt_n, opt_offset) {
                         (Some(s), Some(n), Some(offset)) => {
-                            Some(s.as_ref().gather_every(n as usize, offset as usize))
+                            Ok(Some(s.as_ref().gather_every(n as usize, offset as usize)?))
                         },
-                        _ => None,
-                    }
-                }),
+                        _ => Ok(None),
+                    },
+                )?,
             _ => {
                 polars_bail!(ComputeError: "The lengths of `n` and `offset` should be 1 or equal to the length of list.")
             },
@@ -433,6 +453,12 @@ pub trait ListNameSpaceImpl: AsList {
     #[cfg(feature = "list_gather")]
     fn lst_gather(&self, idx: &Series, null_on_oob: bool) -> PolarsResult<Series> {
         let list_ca = self.as_list();
+        let idx_ca = idx.list()?;
+
+        polars_ensure!(
+            idx_ca.inner_dtype().is_integer(),
+            ComputeError: "cannot use dtype `{}` as an index", idx_ca.inner_dtype()
+        );
 
         let index_typed_index = |idx: &Series| {
             let idx = idx.cast(&IDX_DTYPE).unwrap();
@@ -454,10 +480,66 @@ pub trait ListNameSpaceImpl: AsList {
             }
         };
 
-        use DataType::*;
-        match idx.dtype() {
-            List(boxed_dt) if boxed_dt.is_integer() => {
-                let idx_ca = idx.list().unwrap();
+        match (list_ca.len(), idx_ca.len()) {
+            (1, _) => {
+                let mut out = if list_ca.has_nulls() {
+                    ListChunked::full_null_with_dtype(
+                        PlSmallStr::EMPTY,
+                        idx.len(),
+                        list_ca.inner_dtype(),
+                    )
+                } else {
+                    let s = list_ca.explode(false)?;
+                    idx_ca
+                        .into_iter()
+                        .map(|opt_idx| {
+                            opt_idx
+                                .map(|idx| take_series(&s, idx, null_on_oob))
+                                .transpose()
+                        })
+                        .collect::<PolarsResult<ListChunked>>()?
+                };
+                out.rename(list_ca.name().clone());
+                Ok(out.into_series())
+            },
+            (_, 1) => {
+                let idx_ca = idx_ca.explode(false)?;
+
+                use DataType as D;
+                match idx_ca.dtype() {
+                    D::UInt32 | D::UInt64 => index_typed_index(&idx_ca),
+                    dt if dt.is_signed_integer() => {
+                        if let Some(min) = idx_ca.min::<i64>().unwrap() {
+                            if min >= 0 {
+                                index_typed_index(&idx_ca)
+                            } else {
+                                let mut out = {
+                                    list_ca
+                                        .amortized_iter()
+                                        .map(|opt_s| {
+                                            opt_s
+                                                .map(|s| {
+                                                    take_series(
+                                                        s.as_ref(),
+                                                        idx_ca.clone(),
+                                                        null_on_oob,
+                                                    )
+                                                })
+                                                .transpose()
+                                        })
+                                        .collect::<PolarsResult<ListChunked>>()?
+                                };
+                                out.rename(list_ca.name().clone());
+                                Ok(out.into_series())
+                            }
+                        } else {
+                            polars_bail!(ComputeError: "all indices are null");
+                        }
+                    },
+                    dt => polars_bail!(ComputeError: "cannot use dtype `{dt}` as an index"),
+                }
+            },
+            (a, b) if a == b => {
                 let mut out = {
                     list_ca
                         .amortized_iter()
@@ -476,33 +558,9 @@ pub trait ListNameSpaceImpl: AsList {
                         .collect::<PolarsResult<ListChunked>>()?
                 };
                 out.rename(list_ca.name().clone());
-
                 Ok(out.into_series())
             },
-            UInt32 | UInt64 => index_typed_index(idx),
-            dt if dt.is_signed_integer() => {
-                if let Some(min) = idx.min::<i64>().unwrap() {
-                    if min >= 0 {
-                        index_typed_index(idx)
-                    } else {
-                        let mut out = {
-                            list_ca
-                                .amortized_iter()
-                                .map(|opt_s| {
-                                    opt_s
-                                        .map(|s| take_series(s.as_ref(), idx.clone(), null_on_oob))
-                                        .transpose()
-                                })
-                                .collect::<PolarsResult<ListChunked>>()?
-                        };
-                        out.rename(list_ca.name().clone());
-                        Ok(out.into_series())
-                    }
-                } else {
-                    polars_bail!(ComputeError: "all indices are null");
-                }
-            },
-            dt => polars_bail!(ComputeError: "cannot use dtype `{}` as an index", dt),
+            (a, b) => polars_bail!(length_mismatch = "list.gather", a, b),
         }
     }
 
@@ -521,11 +579,28 @@ pub trait ListNameSpaceImpl: AsList {
         shuffle: bool,
         seed: Option<u64>,
     ) -> PolarsResult<ListChunked> {
+        use std::borrow::Cow;
+
         let ca = self.as_list();
 
         let n_s = n.cast(&IDX_DTYPE)?;
         let n = n_s.idx()?;
 
+        polars_ensure!(
+            ca.len() == n.len() || ca.len() == 1 || n.len() == 1,
+            length_mismatch = "list.sample(n)",
+            ca.len(),
+            n.len()
+        );
+
+        // Broadcast `self`
+        let mut ca = Cow::Borrowed(ca);
+        if ca.len() == 1 && n.len() != 1 {
+            // Optimize: Don't broadcast and instead have a special path.
+            ca = Cow::Owned(ca.new_from_index(0, n.len()));
+        }
+        let ca = ca.as_ref();
+
         let out = match n.len() {
             1 => {
                 if let Some(n) = n.get(0) {
@@ -560,11 +635,28 @@ pub trait ListNameSpaceImpl: AsList {
         shuffle: bool,
         seed: Option<u64>,
     ) -> PolarsResult<ListChunked> {
+        use std::borrow::Cow;
+
         let ca = self.as_list();
 
         let fraction_s = fraction.cast(&DataType::Float64)?;
         let fraction = fraction_s.f64()?;
 
+        polars_ensure!(
+            ca.len() == fraction.len() || ca.len() == 1 || fraction.len() == 1,
+            length_mismatch = "list.sample(fraction)",
+            ca.len(),
+            fraction.len()
+        );
+
+        // Broadcast `self`
+        let mut ca = Cow::Borrowed(ca);
+        if ca.len() == 1 && fraction.len() != 1 {
+            // Optimize: Don't broadcast and instead have a special path.
+            ca = Cow::Owned(ca.new_from_index(0, fraction.len()));
+        }
+        let ca = ca.as_ref();
+
         let out = match fraction.len() {
             1 => {
                 if let Some(fraction) = fraction.get(0) {
