@@ -9,7 +9,9 @@ use crate::{map, map_as_slice, wrap};
 pub enum ListFunction {
     Concat,
     #[cfg(feature = "is_in")]
-    Contains,
+    Contains {
+        nulls_equal: bool,
+    },
     #[cfg(feature = "list_drop_nulls")]
     DropNulls,
     #[cfg(feature = "list_sample")]
@@ -66,7 +68,7 @@ impl ListFunction {
         match self {
             Concat => mapper.map_to_list_supertype(),
             #[cfg(feature = "is_in")]
-            Contains => mapper.with_dtype(DataType::Boolean),
+            Contains { nulls_equal: _ } => mapper.with_dtype(DataType::Boolean),
             #[cfg(feature = "list_drop_nulls")]
             DropNulls => mapper.with_same_dtype(),
             #[cfg(feature = "list_sample")]
@@ -83,14 +85,28 @@ impl ListFunction {
             Sum => mapper.nested_sum_type(),
             Min => mapper.map_to_list_and_array_inner_dtype(),
             Max => mapper.map_to_list_and_array_inner_dtype(),
-            Mean => mapper.with_dtype(DataType::Float64),
-            Median => mapper.map_to_float_dtype(),
+            Mean => mapper.nested_mean_median_type(),
+            Median => mapper.nested_mean_median_type(),
             Std(_) => mapper.map_to_float_dtype(), // Need to also have this sometimes marked as float32 or duration..
             Var(_) => mapper.map_to_float_dtype(),
             ArgMin => mapper.with_dtype(IDX_DTYPE),
             ArgMax => mapper.with_dtype(IDX_DTYPE),
             #[cfg(feature = "diff")]
-            Diff { .. } => mapper.with_same_dtype(),
+            Diff { .. } => mapper.map_dtype(|dt| {
+                let inner_dt = match dt.inner_dtype().unwrap() {
+                    #[cfg(feature = "dtype-datetime")]
+                    DataType::Datetime(tu, _) => DataType::Duration(*tu),
+                    #[cfg(feature = "dtype-date")]
+                    DataType::Date => DataType::Duration(TimeUnit::Milliseconds),
+                    #[cfg(feature = "dtype-time")]
+                    DataType::Time => DataType::Duration(TimeUnit::Nanoseconds),
+                    DataType::UInt64 | DataType::UInt32 => DataType::Int64,
+                    DataType::UInt16 => DataType::Int32,
+                    DataType::UInt8 => DataType::Int16,
+                    inner_dt => inner_dt.clone(),
+                };
+                DataType::List(Box::new(inner_dt))
+            }),
             Sort(_) => mapper.with_same_dtype(),
             Reverse => mapper.with_same_dtype(),
             Unique(_) => mapper.with_same_dtype(),
@@ -109,6 +125,59 @@ impl ListFunction {
             ToStruct(args) => mapper.try_map_dtype(|x| args.get_output_dtype(x)),
         }
     }
+
+    pub fn function_options(&self) -> FunctionOptions {
+        use ListFunction as L;
+        match self {
+            L::Concat => FunctionOptions::elementwise(),
+            #[cfg(feature = "is_in")]
+            L::Contains { nulls_equal: _ } => FunctionOptions::elementwise(),
+            #[cfg(feature = "list_sample")]
+            L::Sample { .. } => FunctionOptions::elementwise(),
+            #[cfg(feature = "list_gather")]
+            L::Gather(_) => FunctionOptions::elementwise(),
+            #[cfg(feature = "list_gather")]
+            L::GatherEvery => FunctionOptions::elementwise(),
+            #[cfg(feature = "list_sets")]
+            L::SetOperation(_) => FunctionOptions::elementwise()
+                .with_casting_rules(CastingRules::Supertype(SuperTypeOptions {
+                    flags: SuperTypeFlags::default() | SuperTypeFlags::ALLOW_IMPLODE_LIST,
+                }))
+                .with_flags(|f| f & !FunctionFlags::RETURNS_SCALAR),
+            #[cfg(feature = "diff")]
+            L::Diff { .. } => FunctionOptions::elementwise(),
+            #[cfg(feature = "list_drop_nulls")]
+            L::DropNulls => FunctionOptions::elementwise(),
+            #[cfg(feature = "list_count")]
+            L::CountMatches => FunctionOptions::elementwise(),
+            L::Sum
+            | L::Slice
+            | L::Shift
+            | L::Get(_)
+            | L::Length
+            | L::Max
+            | L::Min
+            | L::Mean
+            | L::Median
+            | L::Std(_)
+            | L::Var(_)
+            | L::ArgMin
+            | L::ArgMax
+            | L::Sort(_)
+            | L::Reverse
+            | L::Unique(_)
+            | L::Join(_)
+            | L::NUnique => FunctionOptions::elementwise(),
+            #[cfg(feature = "list_any_all")]
+            L::Any | L::All => FunctionOptions::elementwise(),
+            #[cfg(feature = "dtype-array")]
+            L::ToArray(_) => FunctionOptions::elementwise(),
+            #[cfg(feature = "list_to_struct")]
+            L::ToStruct(ListToStructArgs::FixedWidth(_)) => FunctionOptions::elementwise(),
+            #[cfg(feature = "list_to_struct")]
+            L::ToStruct(ListToStructArgs::InferWidth { .. }) => FunctionOptions::groupwise(),
+        }
+    }
 }
 
 #[cfg(feature = "dtype-array")]
@@ -127,7 +196,7 @@ impl Display for ListFunction {
         let name = match self {
             Concat => "concat",
             #[cfg(feature = "is_in")]
-            Contains => "contains",
+            Contains { nulls_equal: _ } => "contains",
             #[cfg(feature = "list_drop_nulls")]
             DropNulls => "drop_nulls",
             #[cfg(feature = "list_sample")]
@@ -191,7 +260,7 @@ impl From<ListFunction> for SpecialEq<Arc<dyn ColumnsUdf>> {
         match func {
             Concat => wrap!(concat),
             #[cfg(feature = "is_in")]
-            Contains => wrap!(contains),
+            Contains { nulls_equal } => map_as_slice!(contains, nulls_equal),
             #[cfg(feature = "list_drop_nulls")]
             DropNulls => map!(drop_nulls),
             #[cfg(feature = "list_sample")]
@@ -248,18 +317,19 @@ impl From<ListFunction> for SpecialEq<Arc<dyn ColumnsUdf>> {
 }
 
 #[cfg(feature = "is_in")]
-pub(super) fn contains(args: &mut [Column]) -> PolarsResult<Option<Column>> {
+pub(super) fn contains(args: &mut [Column], nulls_equal: bool) -> PolarsResult<Column> {
     let list = &args[0];
     let item = &args[1];
     polars_ensure!(matches!(list.dtype(), DataType::List(_)),
         SchemaMismatch: "invalid series dtype: expected `List`, got `{}`", list.dtype(),
     );
-    polars_ops::prelude::is_in(item.as_materialized_series(), list.as_materialized_series()).map(
-        |mut ca| {
-            ca.rename(list.name().clone());
-            Some(ca.into_column())
-        },
-    )
+    let mut ca = polars_ops::prelude::is_in(
+        item.as_materialized_series(),
+        list.as_materialized_series(),
+        nulls_equal,
+    )?;
+    ca.rename(list.name().clone());
+    Ok(ca.into_column())
 }
 
 #[cfg(feature = "list_drop_nulls")]
@@ -400,6 +470,7 @@ pub(super) fn concat(s: &mut [Column]) -> PolarsResult<Option<Column>> {
     let mut first = std::mem::take(&mut s[0]);
     let other = &s[1..];
 
+    // TODO! don't auto cast here, but implode beforehand.
     let mut first_ca = match first.try_list() {
         Some(ca) => ca,
         None => {
@@ -440,8 +511,8 @@ pub(super) fn get(s: &mut [Column], null_on_oob: bool) -> PolarsResult<Option<Co
             }
         },
         len if len == ca.len() => {
-            let ca = ca.rechunk();
-            let arr = ca.downcast_iter().next().unwrap();
+            let tmp = ca.rechunk();
+            let arr = tmp.downcast_as_array();
             let offsets = arr.offsets().as_slice();
             let take_by = if ca.null_count() == 0 {
                 index
@@ -497,6 +568,48 @@ pub(super) fn get(s: &mut [Column], null_on_oob: bool) -> PolarsResult<Option<Co
                 .map(Column::from)
                 .map(Some)
         },
+        _ if ca.len() == 1 => {
+            if ca.null_count() > 0 {
+                return Ok(Some(Column::full_null(
+                    ca.name().clone(),
+                    index.len(),
+                    ca.inner_dtype(),
+                )));
+            }
+            let tmp = ca.rechunk();
+            let arr = tmp.downcast_as_array();
+            let offsets = arr.offsets().as_slice();
+            let start = offsets[0];
+            let end = offsets[1];
+            let out_of_bounds = |offset| offset >= end || offset < start || start == end;
+            let take_by: IdxCa = index
+                .iter()
+                .map(|opt_idx| match opt_idx {
+                    Some(idx) => {
+                        let offset = if idx >= 0 { start + idx } else { end + idx };
+                        if out_of_bounds(offset) {
+                            if null_on_oob {
+                                Ok(None)
+                            } else {
+                                polars_bail!(ComputeError: "get index is out of bounds");
+                            }
+                        } else {
+                            let Ok(offset) = IdxSize::try_from(offset) else {
+                                polars_bail!(ComputeError: "get index is out of bounds");
+                            };
+                            Ok(Some(offset))
+                        }
+                    },
+                    None => Ok(None),
+                })
+                .collect::<Result<IdxCa, _>>()?;
+
+            let s = Series::try_from((ca.name().clone(), arr.values().clone())).unwrap();
+            unsafe { s.take_unchecked(&take_by) }
+                .cast(ca.inner_dtype())
+                .map(Column::from)
+                .map(Some)
+        },
         len => polars_bail!(
             ComputeError:
             "`list.get` expression got an index array of length {} while the list has {} elements",
