@@ -1,8 +1,10 @@
 use arrow::datatypes::ArrowSchemaRef;
 use either::Either;
 use expr_expansion::{is_regex_projection, rewrite_projections};
-use hive::{hive_partitions_from_paths, HivePartitions};
+use hive::hive_partitions_from_paths;
+use polars_core::chunked_array::cast::CastOptions;
 
+use super::convert_utils::SplitPredicates;
 use super::stack_opt::ConversionOptimizer;
 use super::*;
 use crate::plans::conversion::expr_expansion::expand_selectors;
@@ -76,29 +78,37 @@ pub fn to_alp(
         lp_arena,
         conversion_optimizer,
         opt_flags,
+        nodes_scratch: &mut unitvec![],
+        pushdown_maintain_errors: optimizer::pushdown_maintain_errors(),
     };
 
     match to_alp_impl(lp, &mut ctxt) {
         Ok(out) => Ok(out),
         Err(err) => {
-            if let Some(ir_until_then) = lp_arena.last_node() {
-                let node_name = if let PolarsError::Context { msg, .. } = &err {
-                    msg
-                } else {
-                    "THIS_NODE"
-                };
-                let plan = IRPlan::new(
-                    ir_until_then,
-                    std::mem::take(lp_arena),
-                    std::mem::take(expr_arena),
-                );
-                let location = format!("{}", plan.display());
-                Err(err.wrap_msg(|msg| {
-                    format!("{msg}\n\nResolved plan until failure:\n\n\t---> FAILED HERE RESOLVING {node_name} <---\n{location}")
-                }))
+            if opt_flags.contains(OptFlags::EAGER) {
+                // If we dispatched to the lazy engine from the eager API, we don't want to resolve
+                // where in the query plan it went wrong. It is clear from the backtrace anyway.
+                return Err(err.remove_context());
+            };
+
+            let Some(ir_until_then) = lp_arena.last_node() else {
+                return Err(err);
+            };
+
+            let node_name = if let PolarsError::Context { msg, .. } = &err {
+                msg
             } else {
-                Err(err)
-            }
+                "THIS_NODE"
+            };
+            let plan = IRPlan::new(
+                ir_until_then,
+                std::mem::take(lp_arena),
+                std::mem::take(expr_arena),
+            );
+            let location = format!("{}", plan.display());
+            Err(err.wrap_msg(|msg| {
+                format!("{msg}\n\nResolved plan until failure:\n\n\t---> FAILED HERE RESOLVING {node_name} <---\n{location}")
+            }))
         },
     }
 }
@@ -108,6 +118,8 @@ pub(super) struct DslConversionContext<'a> {
     pub(super) lp_arena: &'a mut Arena<IR>,
     pub(super) conversion_optimizer: ConversionOptimizer,
     pub(super) opt_flags: &'a mut OptFlags,
+    pub(super) nodes_scratch: &'a mut UnitVec<Node>,
+    pub(super) pushdown_maintain_errors: bool,
 }
 
 pub(super) fn run_conversion(
@@ -117,7 +129,7 @@ pub(super) fn run_conversion(
 ) -> PolarsResult<Node> {
     let lp_node = ctxt.lp_arena.add(lp);
     ctxt.conversion_optimizer
-        .coerce_types(ctxt.expr_arena, ctxt.lp_arena, lp_node)
+        .optimize_exprs(ctxt.expr_arena, ctxt.lp_arena, lp_node)
         .map_err(|e| e.context(format!("'{name}' failed").into()))?;
 
     Ok(lp_node)
@@ -134,7 +146,7 @@ pub fn to_alp_impl(lp: DslPlan, ctxt: &mut DslConversionContext) -> PolarsResult
         DslPlan::Scan {
             sources,
             file_info,
-            file_options,
+            unified_scan_args: mut unified_scan_args_box,
             scan_type,
             cached_ir,
         } => {
@@ -143,14 +155,17 @@ pub fn to_alp_impl(lp: DslPlan, ctxt: &mut DslConversionContext) -> PolarsResult
             let mut cached_ir = cached_ir.lock().unwrap();
 
             if cached_ir.is_none() {
-                let mut file_options = file_options.clone();
+                let cloud_options = unified_scan_args_box.cloud_options.clone();
+                let cloud_options = cloud_options.as_ref();
+
+                let unified_scan_args = unified_scan_args_box.as_mut();
                 let mut scan_type = scan_type.clone();
 
-                if let Some(hive_schema) = file_options.hive_options.schema.as_deref() {
-                    match file_options.hive_options.enabled {
+                if let Some(hive_schema) = unified_scan_args.hive_options.schema.as_deref() {
+                    match unified_scan_args.hive_options.enabled {
                         // Enable hive_partitioning if it is unspecified but a non-empty hive_schema given
                         None if !hive_schema.is_empty() => {
-                            file_options.hive_options.enabled = Some(true)
+                            unified_scan_args.hive_options.enabled = Some(true)
                         },
                         // hive_partitioning was explicitly disabled
                         Some(false) => polars_bail!(
@@ -161,31 +176,34 @@ pub fn to_alp_impl(lp: DslPlan, ctxt: &mut DslConversionContext) -> PolarsResult
                     }
                 }
 
-                let sources = match &scan_type {
-                    #[cfg(feature = "parquet")]
-                    FileScan::Parquet { cloud_options, .. } => sources
-                        .expand_paths_with_hive_update(&mut file_options, cloud_options.as_ref())?,
-                    #[cfg(feature = "ipc")]
-                    FileScan::Ipc { cloud_options, .. } => sources
-                        .expand_paths_with_hive_update(&mut file_options, cloud_options.as_ref())?,
-                    #[cfg(feature = "csv")]
-                    FileScan::Csv { cloud_options, .. } => {
-                        sources.expand_paths(&file_options, cloud_options.as_ref())?
-                    },
-                    #[cfg(feature = "json")]
-                    FileScan::NDJson { cloud_options, .. } => {
-                        sources.expand_paths(&file_options, cloud_options.as_ref())?
-                    },
-                    FileScan::Anonymous { .. } => sources,
-                };
+                let sources =
+                    match &*scan_type {
+                        #[cfg(feature = "parquet")]
+                        FileScan::Parquet { .. } => sources
+                            .expand_paths_with_hive_update(unified_scan_args, cloud_options)?,
+                        #[cfg(feature = "ipc")]
+                        FileScan::Ipc { .. } => sources
+                            .expand_paths_with_hive_update(unified_scan_args, cloud_options)?,
+                        #[cfg(feature = "csv")]
+                        FileScan::Csv { .. } => {
+                            sources.expand_paths(unified_scan_args, cloud_options)?
+                        },
+                        #[cfg(feature = "json")]
+                        FileScan::NDJson { .. } => {
+                            sources.expand_paths(unified_scan_args, cloud_options)?
+                        },
+                        #[cfg(feature = "python")]
+                        FileScan::PythonDataset { .. } => {
+                            // There are a lot of places that short-circuit if the paths is empty,
+                            // so we just give a dummy path here.
+                            ScanSources::Paths(Arc::from(["dummy".into()]))
+                        },
+                        FileScan::Anonymous { .. } => sources,
+                    };
 
-                let mut file_info = match &mut scan_type {
+                let mut file_info = match &mut *scan_type {
                     #[cfg(feature = "parquet")]
-                    FileScan::Parquet {
-                        options,
-                        cloud_options,
-                        metadata,
-                    } => {
+                    FileScan::Parquet { options, metadata } => {
                         if let Some(schema) = &options.schema {
                             // We were passed a schema, we don't have to call `parquet_file_info`,
                             // but this does mean we don't have `row_estimation` and `first_metadata`.
@@ -199,8 +217,8 @@ pub fn to_alp_impl(lp: DslPlan, ctxt: &mut DslConversionContext) -> PolarsResult
                         } else {
                             let (file_info, md) = scans::parquet_file_info(
                                 &sources,
-                                &file_options,
-                                cloud_options.as_ref(),
+                                unified_scan_args.row_index.as_ref(),
+                                cloud_options,
                             )
                             .map_err(|e| e.context(failed_here!(parquet scan)))?;
 
@@ -209,51 +227,67 @@ pub fn to_alp_impl(lp: DslPlan, ctxt: &mut DslConversionContext) -> PolarsResult
                         }
                     },
                     #[cfg(feature = "ipc")]
-                    FileScan::Ipc {
-                        cloud_options,
-                        metadata,
-                        ..
-                    } => {
-                        let (file_info, md) =
-                            scans::ipc_file_info(&sources, &file_options, cloud_options.as_ref())
-                                .map_err(|e| e.context(failed_here!(ipc scan)))?;
+                    FileScan::Ipc { metadata, .. } => {
+                        let (file_info, md) = scans::ipc_file_info(
+                            &sources,
+                            unified_scan_args.row_index.as_ref(),
+                            cloud_options,
+                        )
+                        .map_err(|e| e.context(failed_here!(ipc scan)))?;
                         *metadata = Some(Arc::new(md));
                         file_info
                     },
                     #[cfg(feature = "csv")]
-                    FileScan::Csv {
-                        options,
-                        cloud_options,
-                    } => scans::csv_file_info(
-                        &sources,
-                        &file_options,
-                        options,
-                        cloud_options.as_ref(),
-                    )
-                    .map_err(|e| e.context(failed_here!(csv scan)))?,
+                    FileScan::Csv { options } => {
+                        // TODO: This is a hack. We conditionally set `allow_missing_columns` to
+                        // mimic existing behavior, but this should be taken from a user provided
+                        // parameter instead.
+                        if options.schema.is_some() && options.has_header {
+                            unified_scan_args.missing_columns_policy = MissingColumnsPolicy::Insert;
+                        }
+
+                        scans::csv_file_info(
+                            &sources,
+                            unified_scan_args.row_index.as_ref(),
+                            options,
+                            cloud_options,
+                        )
+                        .map_err(|e| e.context(failed_here!(csv scan)))?
+                    },
                     #[cfg(feature = "json")]
-                    FileScan::NDJson {
+                    FileScan::NDJson { options } => scans::ndjson_file_info(
+                        &sources,
+                        unified_scan_args.row_index.as_ref(),
                         options,
                         cloud_options,
-                    } => scans::ndjson_file_info(
-                        &sources,
-                        &file_options,
-                        options,
-                        cloud_options.as_ref(),
                     )
                     .map_err(|e| e.context(failed_here!(ndjson scan)))?,
+                    #[cfg(feature = "python")]
+                    FileScan::PythonDataset { dataset_object, .. } => {
+                        if crate::dsl::DATASET_PROVIDER_VTABLE.get().is_none() {
+                            polars_bail!(ComputeError: "DATASET_PROVIDER_VTABLE (python) not initialized")
+                        }
+
+                        let schema = dataset_object.schema()?;
+
+                        FileInfo {
+                            schema: schema.clone(),
+                            reader_schema: Some(either::Either::Right(schema)),
+                            row_estimation: (None, usize::MAX),
+                        }
+                    },
                     FileScan::Anonymous { .. } => {
                         file_info.expect("FileInfo should be set for AnonymousScan")
                     },
                 };
 
-                if file_options.hive_options.enabled.is_none() {
+                if unified_scan_args.hive_options.enabled.is_none() {
                     // We expect this to be `Some(_)` after this point. If it hasn't been auto-enabled
                     // we explicitly set it to disabled.
-                    file_options.hive_options.enabled = Some(false);
+                    unified_scan_args.hive_options.enabled = Some(false);
                 }
 
-                let hive_parts = if file_options.hive_options.enabled.unwrap()
+                let hive_parts = if unified_scan_args.hive_options.enabled.unwrap()
                     && file_info.reader_schema.is_some()
                 {
                     let paths = sources.as_paths().ok_or_else(|| {
@@ -265,8 +299,8 @@ pub fn to_alp_impl(lp: DslPlan, ctxt: &mut DslConversionContext) -> PolarsResult
 
                     hive_partitions_from_paths(
                         paths,
-                        file_options.hive_options.hive_start_idx,
-                        file_options.hive_options.schema.clone(),
+                        unified_scan_args.hive_options.hive_start_idx,
+                        unified_scan_args.hive_options.schema.clone(),
                         match file_info.reader_schema.as_ref().unwrap() {
                             Either::Left(v) => {
                                 owned = Some(Schema::from_arrow_schema(v.as_ref()));
@@ -274,36 +308,23 @@ pub fn to_alp_impl(lp: DslPlan, ctxt: &mut DslConversionContext) -> PolarsResult
                             },
                             Either::Right(v) => v.as_ref(),
                         },
-                        file_options.hive_options.try_parse_dates,
+                        unified_scan_args.hive_options.try_parse_dates,
                     )?
                 } else {
                     None
                 };
 
                 if let Some(ref hive_parts) = hive_parts {
-                    let hive_schema = hive_parts[0].schema();
+                    let hive_schema = hive_parts.schema();
                     file_info.update_schema_with_hive_schema(hive_schema.clone());
-                } else if let Some(hive_schema) = file_options.hive_options.schema.clone() {
+                } else if let Some(hive_schema) = unified_scan_args.hive_options.schema.clone() {
                     // We hit here if we are passed the `hive_schema` to `scan_parquet` but end up with an empty file
                     // list during path expansion. In this case we still want to return an empty DataFrame with this
                     // schema.
                     file_info.update_schema_with_hive_schema(hive_schema);
                 }
 
-                file_options.include_file_paths =
-                    file_options.include_file_paths.filter(|_| match scan_type {
-                        #[cfg(feature = "parquet")]
-                        FileScan::Parquet { .. } => true,
-                        #[cfg(feature = "ipc")]
-                        FileScan::Ipc { .. } => true,
-                        #[cfg(feature = "csv")]
-                        FileScan::Csv { .. } => true,
-                        #[cfg(feature = "json")]
-                        FileScan::NDJson { .. } => true,
-                        FileScan::Anonymous { .. } => false,
-                    });
-
-                if let Some(ref file_path_col) = file_options.include_file_paths {
+                if let Some(ref file_path_col) = unified_scan_args.include_file_paths {
                     let schema = Arc::make_mut(&mut file_info.schema);
 
                     if schema.contains(file_path_col) {
@@ -320,29 +341,32 @@ pub fn to_alp_impl(lp: DslPlan, ctxt: &mut DslConversionContext) -> PolarsResult
                     )?;
                 }
 
-                file_options.with_columns = if file_info.reader_schema.is_some() {
+                unified_scan_args.projection = if file_info.reader_schema.is_some() {
                     maybe_init_projection_excluding_hive(
                         file_info.reader_schema.as_ref().unwrap(),
-                        hive_parts.as_ref().map(|x| &x[0]),
+                        hive_parts.as_ref().map(|h| h.schema()),
                     )
                 } else {
                     None
                 };
 
-                if let Some(row_index) = &file_options.row_index {
+                if let Some(row_index) = &unified_scan_args.row_index {
                     let schema = Arc::make_mut(&mut file_info.schema);
                     *schema = schema
                         .new_inserting_at_index(0, row_index.name.clone(), IDX_DTYPE)
                         .unwrap();
                 }
 
-                let ir = if sources.is_empty() && !matches!(scan_type, FileScan::Anonymous { .. }) {
+                let ir = if sources.is_empty() && !matches!(&*scan_type, FileScan::Anonymous { .. })
+                {
                     IR::DataFrameScan {
                         df: Arc::new(DataFrame::empty_with_schema(&file_info.schema)),
                         schema: file_info.schema,
                         output_schema: None,
                     }
                 } else {
+                    let unified_scan_args = unified_scan_args_box;
+
                     IR::Scan {
                         sources,
                         file_info,
@@ -350,7 +374,8 @@ pub fn to_alp_impl(lp: DslPlan, ctxt: &mut DslConversionContext) -> PolarsResult
                         predicate: None,
                         scan_type,
                         output_schema: None,
-                        file_options,
+                        unified_scan_args,
+                        id: Default::default(),
                     }
                 };
 
@@ -360,7 +385,22 @@ pub fn to_alp_impl(lp: DslPlan, ctxt: &mut DslConversionContext) -> PolarsResult
             cached_ir.clone().unwrap()
         },
         #[cfg(feature = "python")]
-        DslPlan::PythonScan { options } => IR::PythonScan { options },
+        DslPlan::PythonScan { mut options } => {
+            let scan_fn = options.scan_fn.take();
+            let schema = options.get_schema()?;
+            IR::PythonScan {
+                options: PythonOptions {
+                    scan_fn,
+                    schema,
+                    python_source: options.python_source,
+                    validate_schema: options.validate_schema,
+                    output_schema: Default::default(),
+                    with_columns: Default::default(),
+                    n_rows: Default::default(),
+                    predicate: Default::default(),
+                },
+            }
+        },
         DslPlan::Union { inputs, args } => {
             let mut inputs = inputs
                 .into_iter()
@@ -416,54 +456,46 @@ pub fn to_alp_impl(lp: DslPlan, ctxt: &mut DslConversionContext) -> PolarsResult
 
             let predicate_ae = to_expr_ir(predicate.clone(), ctxt.expr_arena)?;
 
-            // TODO: We could do better here by using `pushdown_eligibility()`
-            return if ctxt.opt_flags.predicate_pushdown()
-                && permits_filter_pushdown_rec(
-                    ctxt.expr_arena.get(predicate_ae.node()),
+            if ctxt.opt_flags.predicate_pushdown() {
+                ctxt.nodes_scratch.clear();
+
+                if let Some(SplitPredicates { pushable, fallible }) = SplitPredicates::new(
+                    predicate_ae.node(),
                     ctxt.expr_arena,
+                    Some(ctxt.nodes_scratch),
+                    ctxt.pushdown_maintain_errors,
                 ) {
-                // Split expression that are ANDed into multiple Filter nodes as the optimizer can then
-                // push them down independently. Especially if they refer columns from different tables
-                // this will be more performant.
-                // So:
-                // filter[foo == bar & ham == spam]
-                // filter [foo == bar]
-                // filter [ham == spam]
-                let mut predicates = vec![];
+                    let mut update_input = |predicate: Node| -> PolarsResult<()> {
+                        let predicate = ExprIR::from_node(predicate, ctxt.expr_arena);
+                        ctxt.conversion_optimizer
+                            .push_scratch(predicate.node(), ctxt.expr_arena);
+                        let lp = IR::Filter { input, predicate };
+                        input = run_conversion(lp, ctxt, "filter")?;
 
-                let mut stack = vec![predicate_ae.node()];
-                while let Some(n) = stack.pop() {
-                    if let AExpr::BinaryExpr {
-                        left,
-                        op: Operator::And | Operator::LogicalAnd,
-                        right,
-                    } = ctxt.expr_arena.get(n)
-                    {
-                        stack.push(*left);
-                        stack.push(*right);
-                    } else {
-                        predicates.push(n)
+                        Ok(())
+                    };
+
+                    // Pushables first, then fallible.
+
+                    for predicate in pushable {
+                        update_input(predicate)?;
                     }
-                }
 
-                for predicate in predicates {
-                    let predicate = ExprIR::from_node(predicate, ctxt.expr_arena);
-                    ctxt.conversion_optimizer
-                        .push_scratch(predicate.node(), ctxt.expr_arena);
-                    let lp = IR::Filter { input, predicate };
-                    input = run_conversion(lp, ctxt, "filter")?;
-                }
+                    if let Some(node) = fallible {
+                        update_input(node)?;
+                    }
 
-                Ok(input)
-            } else {
-                ctxt.conversion_optimizer
-                    .push_scratch(predicate_ae.node(), ctxt.expr_arena);
-                let lp = IR::Filter {
-                    input,
-                    predicate: predicate_ae,
+                    return Ok(input);
                 };
-                run_conversion(lp, ctxt, "filter")
             };
+
+            ctxt.conversion_optimizer
+                .push_scratch(predicate_ae.node(), ctxt.expr_arena);
+            let lp = IR::Filter {
+                input,
+                predicate: predicate_ae,
+            };
+            return run_conversion(lp, ctxt, "filter");
         },
         DslPlan::Slice { input, offset, len } => {
             let input =
@@ -556,8 +588,8 @@ pub fn to_alp_impl(lp: DslPlan, ctxt: &mut DslConversionContext) -> PolarsResult
                 )
                 .map_err(|e| e.context(failed_here!(sort)))?;
 
-                nulls_last.extend(std::iter::repeat(n).take(exprs.len()));
-                descending.extend(std::iter::repeat(d).take(exprs.len()));
+                nulls_last.extend(std::iter::repeat_n(n, exprs.len()));
+                descending.extend(std::iter::repeat_n(d, exprs.len()));
                 expanded_cols.extend(exprs);
             }
             sort_options.nulls_last = nulls_last;
@@ -565,7 +597,36 @@ pub fn to_alp_impl(lp: DslPlan, ctxt: &mut DslConversionContext) -> PolarsResult
 
             ctxt.conversion_optimizer
                 .fill_scratch(&expanded_cols, ctxt.expr_arena);
-            let by_column = expanded_cols;
+            let mut by_column = expanded_cols;
+
+            // Remove null columns in multi-columns sort
+            if by_column.len() > 1 {
+                let input_schema = ctxt.lp_arena.get(input).schema(ctxt.lp_arena);
+
+                let mut null_columns = vec![];
+
+                for (i, c) in by_column.iter().enumerate() {
+                    if let DataType::Null =
+                        c.dtype(&input_schema, Context::Default, ctxt.expr_arena)?
+                    {
+                        null_columns.push(i);
+                    }
+                }
+                // All null columns, only take one.
+                if null_columns.len() == by_column.len() {
+                    by_column.truncate(1);
+                    sort_options.nulls_last.truncate(1);
+                    sort_options.descending.truncate(1);
+                }
+                // Remove the null columns
+                else if !null_columns.is_empty() {
+                    for i in null_columns.into_iter().rev() {
+                        by_column.remove(i);
+                        sort_options.nulls_last.remove(i);
+                        sort_options.descending.remove(i);
+                    }
+                }
+            };
 
             let lp = IR::Sort {
                 input,
@@ -596,6 +657,11 @@ pub fn to_alp_impl(lp: DslPlan, ctxt: &mut DslConversionContext) -> PolarsResult
             let input =
                 to_alp_impl(owned(input), ctxt).map_err(|e| e.context(failed_here!(group_by)))?;
 
+            // Rolling + group-by sorts the whole table, so remove unneeded columns
+            if ctxt.opt_flags.eager() && options.is_rolling() && !keys.is_empty() {
+                ctxt.opt_flags.insert(OptFlags::PROJECTION_PUSHDOWN)
+            }
+
             let (keys, aggs, schema) = resolve_group_by(
                 input,
                 keys,
@@ -649,7 +715,7 @@ pub fn to_alp_impl(lp: DslPlan, ctxt: &mut DslConversionContext) -> PolarsResult
                 ctxt,
             )
             .map_err(|e| e.context(failed_here!(join)))
-            .map(|t| t.0)
+            .map(|t| t.0);
         },
         DslPlan::HStack {
             input,
@@ -672,6 +738,125 @@ pub fn to_alp_impl(lp: DslPlan, ctxt: &mut DslConversionContext) -> PolarsResult
             };
             return run_conversion(lp, ctxt, "with_columns");
         },
+        DslPlan::MatchToSchema {
+            input,
+            match_schema,
+            per_column,
+            extra_columns,
+        } => {
+            let input =
+                to_alp_impl(owned(input), ctxt).map_err(|e| e.context(failed_here!(unique)))?;
+            let input_schema = ctxt.lp_arena.get(input).schema(ctxt.lp_arena);
+
+            assert_eq!(per_column.len(), match_schema.len());
+
+            if input_schema.as_ref() == &match_schema {
+                return Ok(input);
+            }
+
+            let mut exprs = Vec::with_capacity(match_schema.len());
+            let mut found_missing_columns = Vec::new();
+            let mut used_input_columns = 0;
+
+            for ((column, dtype), per_column) in match_schema.iter().zip(per_column.iter()) {
+                match input_schema.get(column) {
+                    None => match &per_column.missing_columns {
+                        MissingColumnsPolicyOrExpr::Raise => found_missing_columns.push(column),
+                        MissingColumnsPolicyOrExpr::Insert => exprs.push(Expr::Alias(
+                            Arc::new(Expr::Literal(LiteralValue::Scalar(Scalar::null(
+                                dtype.clone(),
+                            )))),
+                            column.clone(),
+                        )),
+                        MissingColumnsPolicyOrExpr::InsertWith(expr) => {
+                            exprs.push(Expr::Alias(Arc::new(expr.clone()), column.clone()))
+                        },
+                    },
+                    Some(input_dtype) if dtype == input_dtype => {
+                        used_input_columns += 1;
+                        exprs.push(Expr::Column(column.clone()))
+                    },
+                    Some(input_dtype) => {
+                        let from_dtype = input_dtype;
+                        let to_dtype = dtype;
+
+                        let policy = CastColumnsPolicy {
+                            integer_upcast: per_column.integer_cast == UpcastOrForbid::Upcast,
+                            float_upcast: per_column.float_cast == UpcastOrForbid::Upcast,
+                            float_downcast: false,
+                            datetime_nanoseconds_downcast: false,
+                            datetime_microseconds_downcast: false,
+                            datetime_convert_timezone: false,
+                            missing_struct_fields: per_column.missing_struct_fields,
+                            extra_struct_fields: per_column.extra_struct_fields,
+                        };
+
+                        let should_cast =
+                            policy.should_cast_column(column, to_dtype, from_dtype)?;
+
+                        let mut expr = Expr::Column(PlSmallStr::from_str(column));
+                        if should_cast {
+                            expr = Expr::Cast {
+                                expr: Arc::new(expr),
+                                dtype: to_dtype.clone(),
+                                options: CastOptions::NonStrict,
+                            };
+                        }
+
+                        used_input_columns += 1;
+                        exprs.push(expr);
+                    },
+                }
+            }
+
+            // Report the error for missing columns
+            if let Some(lst) = found_missing_columns.first() {
+                use std::fmt::Write;
+                let mut formatted = String::new();
+                write!(&mut formatted, "\"{}\"", found_missing_columns[0]).unwrap();
+                for c in &found_missing_columns[1..] {
+                    write!(&mut formatted, ", \"{c}\"").unwrap();
+                }
+
+                write!(&mut formatted, "\"{lst}\"").unwrap();
+                polars_bail!(SchemaMismatch: "missing columns in `match_to_schema`: {formatted}");
+            }
+
+            // Report the error for extra columns
+            if used_input_columns != input_schema.len()
+                && extra_columns == ExtraColumnsPolicy::Raise
+            {
+                let found_extra_columns = input_schema
+                    .iter_names()
+                    .filter(|n| !match_schema.contains(n))
+                    .collect::<Vec<_>>();
+
+                use std::fmt::Write;
+                let mut formatted = String::new();
+                write!(&mut formatted, "\"{}\"", found_extra_columns[0]).unwrap();
+                for c in &found_extra_columns[1..] {
+                    write!(&mut formatted, ", \"{c}\"").unwrap();
+                }
+
+                polars_bail!(SchemaMismatch: "extra columns in `match_to_schema`: {formatted}");
+            }
+
+            let exprs = to_expr_irs(exprs, ctxt.expr_arena)?;
+
+            ctxt.conversion_optimizer
+                .fill_scratch(&exprs, ctxt.expr_arena);
+            let lp = IR::Select {
+                input,
+                expr: exprs,
+                schema: match_schema.clone(),
+                options: ProjectionOptions {
+                    run_parallel: true,
+                    duplicate_check: false,
+                    should_broadcast: true,
+                },
+            };
+            return run_conversion(lp, ctxt, "match_to_schema");
+        },
         DslPlan::Distinct { input, options } => {
             let input =
                 to_alp_impl(owned(input), ctxt).map_err(|e| e.context(failed_here!(unique)))?;
@@ -902,7 +1087,76 @@ pub fn to_alp_impl(lp: DslPlan, ctxt: &mut DslConversionContext) -> PolarsResult
         DslPlan::Sink { input, payload } => {
             let input =
                 to_alp_impl(owned(input), ctxt).map_err(|e| e.context(failed_here!(sink)))?;
-            IR::Sink { input, payload }
+            let payload = match payload {
+                SinkType::Memory => SinkTypeIR::Memory,
+                SinkType::File(f) => SinkTypeIR::File(f),
+                SinkType::Partition(f) => SinkTypeIR::Partition(PartitionSinkTypeIR {
+                    base_path: f.base_path,
+                    file_path_cb: f.file_path_cb,
+                    file_type: f.file_type,
+                    sink_options: f.sink_options,
+                    variant: match f.variant {
+                        PartitionVariant::MaxSize(max_size) => {
+                            PartitionVariantIR::MaxSize(max_size)
+                        },
+                        PartitionVariant::Parted {
+                            key_exprs,
+                            include_key,
+                        } => {
+                            let eirs = to_expr_irs(key_exprs, ctxt.expr_arena)?;
+                            ctxt.conversion_optimizer
+                                .fill_scratch(&eirs, ctxt.expr_arena);
+
+                            PartitionVariantIR::Parted {
+                                key_exprs: eirs,
+                                include_key,
+                            }
+                        },
+                        PartitionVariant::ByKey {
+                            key_exprs,
+                            include_key,
+                        } => {
+                            let eirs = to_expr_irs(key_exprs, ctxt.expr_arena)?;
+                            ctxt.conversion_optimizer
+                                .fill_scratch(&eirs, ctxt.expr_arena);
+
+                            PartitionVariantIR::ByKey {
+                                key_exprs: eirs,
+                                include_key,
+                            }
+                        },
+                    },
+                    cloud_options: f.cloud_options,
+                }),
+            };
+
+            let lp = IR::Sink { input, payload };
+            return run_conversion(lp, ctxt, "sink");
+        },
+        DslPlan::SinkMultiple { inputs } => {
+            let inputs = inputs
+                .into_iter()
+                .map(|lp| to_alp_impl(lp, ctxt))
+                .collect::<PolarsResult<Vec<_>>>()
+                .map_err(|e| e.context(failed_here!(vertical concat)))?;
+            IR::SinkMultiple { inputs }
+        },
+        #[cfg(feature = "merge_sorted")]
+        DslPlan::MergeSorted {
+            input_left,
+            input_right,
+            key,
+        } => {
+            let input_left = to_alp_impl(owned(input_left), ctxt)
+                .map_err(|e| e.context(failed_here!(merge_sorted)))?;
+            let input_right = to_alp_impl(owned(input_right), ctxt)
+                .map_err(|e| e.context(failed_here!(merge_sorted)))?;
+
+            IR::MergeSorted {
+                input_left,
+                input_right,
+                key,
+            }
         },
         DslPlan::IR { node, dsl, version } => {
             return if node.is_some()
@@ -912,7 +1166,7 @@ pub fn to_alp_impl(lp: DslPlan, ctxt: &mut DslConversionContext) -> PolarsResult
                 Ok(node.unwrap())
             } else {
                 to_alp_impl(owned(dsl), ctxt)
-            }
+            };
         },
     };
     Ok(ctxt.lp_arena.add(v))
@@ -965,11 +1219,15 @@ fn expand_filter(
                 }
 
                 let msg = if cfg!(feature = "python") {
-                    format!("The predicate passed to 'LazyFrame.filter' expanded to multiple expressions: \n\n{expanded}\n\
-                            This is ambiguous. Try to combine the predicates with the 'all' or `any' expression.")
+                    format!(
+                        "The predicate passed to 'LazyFrame.filter' expanded to multiple expressions: \n\n{expanded}\n\
+                            This is ambiguous. Try to combine the predicates with the 'all' or `any' expression."
+                    )
                 } else {
-                    format!("The predicate passed to 'LazyFrame.filter' expanded to multiple expressions: \n\n{expanded}\n\
-                            This is ambiguous. Try to combine the predicates with the 'all_horizontal' or `any_horizontal' expression.")
+                    format!(
+                        "The predicate passed to 'LazyFrame.filter' expanded to multiple expressions: \n\n{expanded}\n\
+                            This is ambiguous. Try to combine the predicates with the 'all_horizontal' or `any_horizontal' expression."
+                    )
                 };
                 polars_bail!(ComputeError: msg)
             },
@@ -1104,12 +1362,11 @@ where
 
 pub(crate) fn maybe_init_projection_excluding_hive(
     reader_schema: &Either<ArrowSchemaRef, SchemaRef>,
-    hive_parts: Option<&HivePartitions>,
+    hive_parts: Option<&SchemaRef>,
 ) -> Option<Arc<[PlSmallStr]>> {
     // Update `with_columns` with a projection so that hive columns aren't loaded from the
     // file
-    let hive_parts = hive_parts?;
-    let hive_schema = hive_parts.schema();
+    let hive_schema = hive_parts?;
 
     match &reader_schema {
         Either::Left(reader_schema) => hive_schema
