@@ -109,8 +109,8 @@ fn expand_regex(
     pattern: &str,
     exclude: &PlHashSet<PlSmallStr>,
 ) -> PolarsResult<()> {
-    let re =
-        regex::Regex::new(pattern).map_err(|e| polars_err!(ComputeError: "invalid regex {}", e))?;
+    let re = polars_utils::regex_cache::compile_regex(pattern)
+        .map_err(|e| polars_err!(ComputeError: "invalid regex {}", e))?;
     for name in schema.iter_names() {
         if re.is_match(name) && !exclude.contains(name.as_str()) {
             let mut new_expr = remove_exclude(expr.clone());
@@ -177,7 +177,7 @@ fn expand_columns(
 ) -> PolarsResult<()> {
     if !expr.into_iter().all(|e| match e {
         // check for invalid expansions such as `col([a, b]) + col([c, d])`
-        Expr::Columns(ref members) => members.as_ref() == names,
+        Expr::Columns(members) => members.as_ref() == names,
         _ => true,
     }) {
         polars_bail!(ComputeError: "expanding more than one `col` is not allowed");
@@ -253,8 +253,10 @@ fn dtypes_match(d1: &DataType, d2: &DataType) -> bool {
         (DataType::Datetime(tu_l, tz_l), DataType::Datetime(tu_r, tz_r)) => {
             tu_l == tu_r
                 && (tz_l == tz_r
-                    || tz_r.is_some() && (tz_l.as_deref().unwrap_or("") == "*")
-                    || tz_l.is_some() && (tz_r.as_deref().unwrap_or("") == "*"))
+                    || match (tz_l, tz_r) {
+                        (Some(l), Some(r)) => TimeZone::eq_wildcard_aware(l, r),
+                        _ => false,
+                    })
         },
         // ...but otherwise require exact match
         _ => d1 == d2,
@@ -331,7 +333,9 @@ fn expand_struct_fields(
     names: &[PlSmallStr],
     exclude: &PlHashSet<PlSmallStr>,
 ) -> PolarsResult<()> {
-    let first_name = names[0].as_ref();
+    let Some(first_name) = names.first() else {
+        return Ok(());
+    };
     if names.len() == 1 && first_name == "*" || is_regex_projection(first_name) {
         let Expr::Function { input, .. } = struct_expr else {
             unreachable!()
@@ -395,7 +399,7 @@ Hint: set 'upper_bound' for 'list.to_struct'.",
         else {
             #[cfg(feature = "regex")]
             {
-                let re = regex::Regex::new(first_name)
+                let re = polars_utils::regex_cache::compile_regex(first_name)
                     .map_err(|e| polars_err!(ComputeError: "invalid regex {}", e))?;
 
                 fields
@@ -565,7 +569,9 @@ fn expand_function_inputs(
             *input = rewrite_projections(core::mem::take(input), schema, &[], opt_flags)?;
             if input.is_empty() && !options.flags.contains(FunctionFlags::ALLOW_EMPTY_INPUTS) {
                 // Needed to visualize the error
-                *input = vec![Expr::Literal(LiteralValue::Null)];
+                *input = vec![Expr::Literal(LiteralValue::Scalar(Scalar::null(
+                    DataType::Null,
+                )))];
                 polars_bail!(InvalidOperation: "expected at least 1 input in {}", e)
             }
             Ok(e)
