@@ -2,7 +2,7 @@ use arrow::legacy::error::PolarsResult;
 use either::Either;
 use polars_core::chunked_array::cast::CastOptions;
 use polars_core::error::feature_gated;
-use polars_core::utils::get_numeric_upcast_supertype_lossless;
+use polars_core::utils::{get_numeric_upcast_supertype_lossless, try_get_supertype};
 use polars_utils::format_pl_smallstr;
 use polars_utils::itertools::Itertools;
 
@@ -73,7 +73,9 @@ pub fn resolve_join(
         if turn_off_coalesce {
             let options = Arc::make_mut(&mut options);
             if matches!(options.args.coalesce, JoinCoalesce::CoalesceColumns) {
-                polars_warn!("coalescing join requested but not all join keys are column references, turning off key coalescing");
+                polars_warn!(
+                    "coalescing join requested but not all join keys are column references, turning off key coalescing"
+                );
             }
             options.args.coalesce = JoinCoalesce::KeepColumns;
         }
@@ -129,26 +131,18 @@ pub fn resolve_join(
     ctxt.conversion_optimizer
         .fill_scratch(&left_on, ctxt.expr_arena);
     ctxt.conversion_optimizer
-        .coerce_types(ctxt.expr_arena, ctxt.lp_arena, input_left)
+        .optimize_exprs(ctxt.expr_arena, ctxt.lp_arena, input_left)
         .map_err(|e| e.context("'join' failed".into()))?;
     ctxt.conversion_optimizer
         .fill_scratch(&right_on, ctxt.expr_arena);
     ctxt.conversion_optimizer
-        .coerce_types(ctxt.expr_arena, ctxt.lp_arena, input_right)
+        .optimize_exprs(ctxt.expr_arena, ctxt.lp_arena, input_right)
         .map_err(|e| e.context("'join' failed".into()))?;
 
     // Re-evaluate because of mutable borrows earlier.
     let schema_left = ctxt.lp_arena.get(input_left).schema(ctxt.lp_arena);
     let schema_right = ctxt.lp_arena.get(input_right).schema(ctxt.lp_arena);
 
-    // Not a closure to avoid borrow issues because we mutate expr_arena as well.
-    macro_rules! get_dtype {
-        ($expr:expr, $schema:expr) => {
-            ctxt.expr_arena
-                .get($expr.node())
-                .get_type($schema, Context::Default, ctxt.expr_arena)
-        };
-    }
     // # Resolve scalars
     //
     // Scalars need to be expanded. We translate them to temporary columns added with
@@ -234,6 +228,15 @@ pub fn resolve_join(
         (schema_left, schema_right)
     };
 
+    // Not a closure to avoid borrow issues because we mutate expr_arena as well.
+    macro_rules! get_dtype {
+        ($expr:expr, $schema:expr) => {
+            ctxt.expr_arena
+                .get($expr.node())
+                .get_type($schema, Context::Default, ctxt.expr_arena)
+        };
+    }
+
     // # Cast lossless
     //
     // If we do a full join and keys are coalesced, the cast keys must be added up front.
@@ -242,22 +245,24 @@ pub fn resolve_join(
     let mut as_with_columns_l = vec![];
     let mut as_with_columns_r = vec![];
     for (lnode, rnode) in left_on.iter_mut().zip(right_on.iter_mut()) {
-        //polars_ensure!(!lnode.is_scalar(&ctxt.expr_arena), InvalidOperation: "joining on scalars is not allowed, consider using 'join_where'");
-        //polars_ensure!(!rnode.is_scalar(&ctxt.expr_arena), InvalidOperation: "joining on scalars is not allowed, consider using 'join_where'");
-
         let ltype = get_dtype!(lnode, &schema_left)?;
         let rtype = get_dtype!(rnode, &schema_right)?;
 
         if let Some(dtype) = get_numeric_upcast_supertype_lossless(&ltype, &rtype) {
+            // We use overflowing cast to allow better optimization as we are casting to a known
+            // lossless supertype.
+            //
+            // We have unique references to these nodes (they are created by this function),
+            // so we can mutate in-place without causing side effects somewhere else.
             let casted_l = ctxt.expr_arena.add(AExpr::Cast {
                 expr: lnode.node(),
                 dtype: dtype.clone(),
-                options: CastOptions::Strict,
+                options: CastOptions::Overflowing,
             });
             let casted_r = ctxt.expr_arena.add(AExpr::Cast {
                 expr: rnode.node(),
                 dtype,
-                options: CastOptions::Strict,
+                options: CastOptions::Overflowing,
             });
 
             if key_cols_coalesced {
@@ -400,37 +405,12 @@ fn resolve_join_where(
     let input_right = to_alp_impl(Arc::unwrap_or_clone(input_right), ctxt)
         .map_err(|e| e.context(failed_here!(join left)))?;
 
-    let schema_left = ctxt.lp_arena.get(input_left).schema(ctxt.lp_arena);
-    let schema_right = ctxt
+    let schema_left = ctxt
         .lp_arena
-        .get(input_right)
+        .get(input_left)
         .schema(ctxt.lp_arena)
         .into_owned();
 
-    for expr in &predicates {
-        fn all_in_schema(
-            schema: &Schema,
-            other: Option<&Schema>,
-            left: &Expr,
-            right: &Expr,
-        ) -> bool {
-            let mut iter =
-                expr_to_leaf_column_names_iter(left).chain(expr_to_leaf_column_names_iter(right));
-            iter.all(|name| {
-                schema.contains(name.as_str()) && other.is_none_or(|s| !s.contains(name.as_str()))
-            })
-        }
-
-        let valid = expr.into_iter().all(|e| match e {
-            Expr::BinaryExpr { left, op, right } if op.is_comparison() => {
-                !(all_in_schema(&schema_left, None, left, right)
-                    || all_in_schema(&schema_right, Some(&schema_left), left, right))
-            },
-            _ => true,
-        });
-        polars_ensure!( valid, InvalidOperation: "'join_where' predicate only refers to columns from a single table")
-    }
-
     let opts = Arc::make_mut(&mut options);
     opts.args.how = JoinType::Cross;
 
@@ -444,14 +424,172 @@ fn resolve_join_where(
         ctxt,
     )?;
 
+    let schema_merged = ctxt
+        .lp_arena
+        .get(last_node)
+        .schema(ctxt.lp_arena)
+        .into_owned();
+
+    // Perform predicate validation.
+    let mut upcast_exprs = Vec::<(Node, DataType)>::new();
     for e in predicates {
-        let predicate = to_expr_ir_ignore_alias(e, ctxt.expr_arena)?;
+        let arena = &mut ctxt.expr_arena;
+        let predicate = to_expr_ir_ignore_alias(e, arena)?;
+        let node = predicate.node();
+
+        // Ensure the predicate dtype output of the root node is Boolean
+        let ae = arena.get(node);
+        let dt_out = ae.to_dtype(&schema_merged, Context::Default, arena)?;
+        polars_ensure!(
+            dt_out == DataType::Boolean,
+            ComputeError: "'join_where' predicates must resolve to boolean"
+        );
+
+        ensure_lossless_binary_comparisons(
+            &node,
+            &schema_left,
+            &schema_merged,
+            arena,
+            &mut upcast_exprs,
+        )?;
+
+        ctxt.conversion_optimizer
+            .push_scratch(predicate.node(), ctxt.expr_arena);
 
         let ir = IR::Filter {
             input: last_node,
             predicate,
         };
+
         last_node = ctxt.lp_arena.add(ir);
     }
+
+    ctxt.conversion_optimizer
+        .optimize_exprs(ctxt.expr_arena, ctxt.lp_arena, last_node)
+        .map_err(|e| e.context("'join_where' failed".into()))?;
+
     Ok((last_node, join_node))
 }
+
+/// Locate nodes that are operands in a binary comparison involving both tables, and ensure that
+/// these nodes are losslessly upcast to a safe dtype.
+fn ensure_lossless_binary_comparisons(
+    node: &Node,
+    schema_left: &Schema,
+    schema_merged: &Schema,
+    expr_arena: &mut Arena<AExpr>,
+    upcast_exprs: &mut Vec<(Node, DataType)>,
+) -> PolarsResult<()> {
+    // let mut upcast_exprs = Vec::<(Node, DataType)>::new();
+    // Ensure that all binary comparisons that use both tables are lossless.
+    build_upcast_node_list(node, schema_left, schema_merged, expr_arena, upcast_exprs)?;
+    // Replace each node with its casted counterpart
+    for (expr, dtype) in upcast_exprs.drain(..) {
+        let old_expr = expr_arena.duplicate(expr);
+        let new_aexpr = AExpr::Cast {
+            expr: old_expr,
+            dtype,
+            options: CastOptions::Overflowing,
+        };
+        expr_arena.replace(expr, new_aexpr);
+    }
+    Ok(())
+}
+
+/// If we are dealing with a binary comparison involving columns from exclusively the left table
+/// on the LHS and the right table on the RHS side, ensure that the cast is lossless.
+/// Expressions involving binaries using either table alone we leave up to the user to verify
+/// that they are valid, as they could theoretically be pushed outside of the join.
+#[recursive]
+fn build_upcast_node_list(
+    node: &Node,
+    schema_left: &Schema,
+    schema_merged: &Schema,
+    expr_arena: &Arena<AExpr>,
+    to_replace: &mut Vec<(Node, DataType)>,
+) -> PolarsResult<ExprOrigin> {
+    let expr_origin = match expr_arena.get(*node) {
+        AExpr::Column(name) => {
+            if schema_left.contains(name) {
+                ExprOrigin::Left
+            } else if schema_merged.contains(name) {
+                ExprOrigin::Right
+            } else {
+                polars_bail!(ColumnNotFound: "{}", name);
+            }
+        },
+        AExpr::Literal(..) => ExprOrigin::None,
+        AExpr::Cast { expr: node, .. } => {
+            build_upcast_node_list(node, schema_left, schema_merged, expr_arena, to_replace)?
+        },
+        AExpr::BinaryExpr {
+            left: left_node,
+            op,
+            right: right_node,
+        } => {
+            // If left and right node has both, ensure the dtypes are valid.
+            let left_origin = build_upcast_node_list(
+                left_node,
+                schema_left,
+                schema_merged,
+                expr_arena,
+                to_replace,
+            )?;
+            let right_origin = build_upcast_node_list(
+                right_node,
+                schema_left,
+                schema_merged,
+                expr_arena,
+                to_replace,
+            )?;
+            // We only update casts during comparisons if the operands are from different tables.
+            if op.is_comparison() {
+                match (left_origin, right_origin) {
+                    (ExprOrigin::Left, ExprOrigin::Right)
+                    | (ExprOrigin::Right, ExprOrigin::Left) => {
+                        // Ensure our dtype casts are lossless
+                        let left = expr_arena.get(*left_node);
+                        let right = expr_arena.get(*right_node);
+                        let dtype_left =
+                            left.to_dtype(schema_merged, Context::Default, expr_arena)?;
+                        let dtype_right =
+                            right.to_dtype(schema_merged, Context::Default, expr_arena)?;
+                        if dtype_left != dtype_right {
+                            // Ensure that we have a lossless cast between the two types.
+                            let dt = if dtype_left.is_primitive_numeric()
+                                || dtype_right.is_primitive_numeric()
+                            {
+                                get_numeric_upcast_supertype_lossless(&dtype_left, &dtype_right)
+                                    .ok_or(PolarsError::SchemaMismatch(
+                                        format!(
+                                            "'join_where' cannot compare {:?} with {:?}",
+                                            dtype_left, dtype_right
+                                        )
+                                        .into(),
+                                    ))
+                            } else {
+                                try_get_supertype(&dtype_left, &dtype_right)
+                            }?;
+
+                            // Store the nodes and their replacements if a cast is required.
+                            let replace_left = dt != dtype_left;
+                            let replace_right = dt != dtype_right;
+                            if replace_left && replace_right {
+                                to_replace.push((*left_node, dt.clone()));
+                                to_replace.push((*right_node, dt));
+                            } else if replace_left {
+                                to_replace.push((*left_node, dt));
+                            } else if replace_right {
+                                to_replace.push((*right_node, dt));
+                            }
+                        }
+                    },
+                    _ => (),
+                }
+            }
+            left_origin | right_origin
+        },
+        _ => ExprOrigin::None,
+    };
+    Ok(expr_origin)
+}
