@@ -1,8 +1,7 @@
 use std::fmt;
 
 use polars_core::error::*;
-#[cfg(feature = "regex")]
-use regex::Regex;
+use polars_utils::{format_list_container_truncated, format_list_truncated};
 
 use crate::constants;
 use crate::plans::ir::IRPlanRef;
@@ -25,24 +24,31 @@ pub struct TreeFmtAExpr<'a>(&'a AExpr);
 impl fmt::Display for TreeFmtAExpr<'_> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         let s = match self.0 {
-            AExpr::Explode(_) => "explode",
+            AExpr::Explode {
+                expr: _,
+                skip_empty: false,
+            } => "explode",
+            AExpr::Explode {
+                expr: _,
+                skip_empty: true,
+            } => "explode(skip_empty)",
             AExpr::Alias(_, name) => return write!(f, "alias({})", name),
             AExpr::Column(name) => return write!(f, "col({})", name),
             AExpr::Literal(lv) => return write!(f, "lit({lv:?})"),
             AExpr::BinaryExpr { op, .. } => return write!(f, "binary: {}", op),
             AExpr::Cast { dtype, options, .. } => {
-                return if options.strict() {
+                return if options.is_strict() {
                     write!(f, "strict cast({})", dtype)
                 } else {
                     write!(f, "cast({})", dtype)
-                }
+                };
             },
             AExpr::Sort { options, .. } => {
                 return write!(
                     f,
                     "sort: {}{}{}",
                     options.descending as u8, options.nulls_last as u8, options.multithreaded as u8
-                )
+                );
             },
             AExpr::Gather { .. } => "gather",
             AExpr::SortBy { sort_options, .. } => {
@@ -63,7 +69,7 @@ impl fmt::Display for TreeFmtAExpr<'_> {
             },
             AExpr::Ternary { .. } => "ternary",
             AExpr::AnonymousFunction { options, .. } => {
-                return write!(f, "anonymous_function: {}", options.fmt_str)
+                return write!(f, "anonymous_function: {}", options.fmt_str);
             },
             AExpr::Function { function, .. } => return write!(f, "function: {function}"),
             AExpr::Window { .. } => "window",
@@ -92,8 +98,10 @@ fn with_header(header: &Option<String>, text: &str) -> String {
 
 #[cfg(feature = "regex")]
 fn multiline_expression(expr: &str) -> std::borrow::Cow<'_, str> {
-    let re = Regex::new(r"([\)\]])(\.[a-z0-9]+\()").unwrap();
-    re.replace_all(expr, "$1\n  $2")
+    polars_utils::regex_cache::cached_regex! {
+        static RE = r"([\)\]])(\.[a-z0-9]+\()";
+    }
+    RE.replace_all(expr, "$1\n  $2")
 }
 
 impl<'a> TreeFmtNode<'a> {
@@ -164,7 +172,7 @@ impl<'a> TreeFmtNode<'a> {
     }
 
     fn node_data(&self) -> TreeFmtNodeData<'_> {
-        use {with_header as wh, TreeFmtNodeContent as C, TreeFmtNodeData as ND};
+        use {TreeFmtNodeContent as C, TreeFmtNodeData as ND, with_header as wh};
 
         let lp = &self.lp;
         let h = &self.h;
@@ -190,22 +198,33 @@ impl<'a> TreeFmtNode<'a> {
                         schema,
                         output_schema,
                         ..
-                    } => ND(
-                        wh(
-                            h,
-                            &format!(
-                                "DF {:?}\nPROJECT {}/{} COLUMNS",
-                                schema.iter_names().take(4).collect::<Vec<_>>(),
-                                if let Some(columns) = output_schema {
-                                    format!("{}", columns.len())
-                                } else {
-                                    "*".to_string()
-                                },
-                                schema.len()
+                    } => {
+                        let (n_columns, projected) = if let Some(schema) = output_schema {
+                            (
+                                format!("{}", schema.len()),
+                                format!(
+                                    ": {};",
+                                    format_list_truncated!(schema.iter_names(), 4, '"')
+                                ),
+                            )
+                        } else {
+                            ("*".to_string(), "".to_string())
+                        };
+                        ND(
+                            wh(
+                                h,
+                                &format!(
+                                    "DF {}\nPROJECT{} {}/{} COLUMNS",
+                                    format_list_truncated!(schema.iter_names(), 4, '"'),
+                                    projected,
+                                    n_columns,
+                                    schema.len()
+                                ),
                             ),
-                        ),
-                        vec![],
-                    ),
+                            vec![],
+                        )
+                    },
+
                     Union { inputs, .. } => ND(
                         wh(
                             h,
@@ -267,9 +286,16 @@ impl<'a> TreeFmtNode<'a> {
                             .collect(),
                     ),
                     GroupBy {
-                        input, keys, aggs, ..
+                        input,
+                        keys,
+                        aggs,
+                        maintain_order,
+                        ..
                     } => ND(
-                        wh(h, "AGGREGATE"),
+                        wh(
+                            h,
+                            &format!("AGGREGATE[maintain_order: {}]", *maintain_order),
+                        ),
                         aggs.iter()
                             .map(|expr| self.expr_node(Some("expression:".to_string()), expr))
                             .chain(keys.iter().map(|expr| {
@@ -341,12 +367,21 @@ impl<'a> TreeFmtNode<'a> {
                         wh(
                             h,
                             match payload {
-                                SinkType::Memory => "SINK (memory)",
-                                SinkType::File { .. } => "SINK (file)",
+                                SinkTypeIR::Memory => "SINK (memory)",
+                                SinkTypeIR::File { .. } => "SINK (file)",
+                                SinkTypeIR::Partition { .. } => "SINK (partition)",
                             },
                         ),
                         vec![self.lp_node(None, *input)],
                     ),
+                    SinkMultiple { inputs } => ND(
+                        wh(h, "SINK_MULTIPLE"),
+                        inputs
+                            .iter()
+                            .enumerate()
+                            .map(|(i, lp_root)| self.lp_node(Some(format!("PLAN {i}:")), *lp_root))
+                            .collect(),
+                    ),
                     SimpleProjection { input, columns } => {
                         let num_columns = columns.as_ref().len();
                         let total_columns = lp.lp_arena.get(*input).schema(lp.lp_arena).len();
@@ -360,6 +395,18 @@ impl<'a> TreeFmtNode<'a> {
                             vec![self.lp_node(None, *input)],
                         )
                     },
+                    #[cfg(feature = "merge_sorted")]
+                    MergeSorted {
+                        input_left,
+                        input_right,
+                        key,
+                    } => ND(
+                        wh(h, &format!("MERGE SORTED ON '{key}")),
+                        [self.lp_node(Some("LEFT PLAN:".to_string()), *input_left)]
+                            .into_iter()
+                            .chain([self.lp_node(Some("RIGHT PLAN:".to_string()), *input_right)])
+                            .collect(),
+                    ),
                     Invalid => ND(wh(h, "INVALID"), vec![]),
                 }
             },
@@ -798,11 +845,7 @@ impl From<TreeView<'_>> for Canvas {
         }
 
         fn even_odd(a: usize, b: usize) -> usize {
-            if a % 2 == 0 && b % 2 == 1 {
-                1
-            } else {
-                0
-            }
+            if a % 2 == 0 && b % 2 == 1 { 1 } else { 0 }
         }
 
         for (i, row) in value.matrix.iter().enumerate() {
