@@ -1,15 +1,16 @@
+use std::hash::BuildHasher;
+
 use hashbrown::hash_map::RawEntryMut;
+use polars_core::CHEAP_SERIES_HASH_LIMIT;
+use polars_utils::aliases::PlFixedStateQuality;
 use polars_utils::format_pl_smallstr;
+use polars_utils::hashing::_boost_hash_combine;
 use polars_utils::vec::CapacityByFactor;
 
 use super::*;
 use crate::constants::CSE_REPLACED;
 use crate::prelude::visitor::AexprNode;
 
-const SERIES_LIMIT: usize = 1000;
-
-use polars_utils::hashing::_boost_hash_combine;
-
 #[derive(Debug, Clone)]
 struct ProjectionExprs {
     expr: Vec<ExprIR>,
@@ -45,7 +46,7 @@ impl ProjectionExprs {
 pub(super) struct Identifier {
     inner: Option<u64>,
     last_node: Option<AexprNode>,
-    hb: PlRandomState,
+    hb: PlFixedStateQuality,
 }
 
 impl Identifier {
@@ -53,7 +54,7 @@ impl Identifier {
         Self {
             inner: None,
             last_node: None,
-            hb: PlRandomState::with_seed(0),
+            hb: PlFixedStateQuality::with_seed(0),
         }
     }
 
@@ -102,7 +103,7 @@ impl Identifier {
         Self {
             inner,
             last_node: Some(*ae),
-            hb: self.hb.clone(),
+            hb: self.hb,
         }
     }
 }
@@ -148,6 +149,69 @@ impl<V> IdentifierMap<V> {
     }
 }
 
+/// Merges identical expressions into identical IDs.
+///
+/// Does no analysis whether this leads to legal substitutions.
+#[derive(Default)]
+pub struct NaiveExprMerger {
+    node_to_uniq_id: PlHashMap<Node, u32>,
+    uniq_id_to_node: Vec<Node>,
+    identifier_to_uniq_id: IdentifierMap<u32>,
+    arg_stack: Vec<Option<Identifier>>,
+}
+
+impl NaiveExprMerger {
+    pub fn add_expr(&mut self, node: Node, arena: &Arena<AExpr>) {
+        let node = AexprNode::new(node);
+        node.visit(self, arena).unwrap();
+    }
+
+    pub fn get_uniq_id(&self, node: Node) -> Option<u32> {
+        self.node_to_uniq_id.get(&node).copied()
+    }
+
+    pub fn get_node(&self, uniq_id: u32) -> Option<Node> {
+        self.uniq_id_to_node.get(uniq_id as usize).copied()
+    }
+}
+
+impl Visitor for NaiveExprMerger {
+    type Node = AexprNode;
+    type Arena = Arena<AExpr>;
+
+    fn pre_visit(
+        &mut self,
+        _node: &Self::Node,
+        _arena: &Self::Arena,
+    ) -> PolarsResult<VisitRecursion> {
+        self.arg_stack.push(None);
+        Ok(VisitRecursion::Continue)
+    }
+
+    fn post_visit(
+        &mut self,
+        node: &Self::Node,
+        arena: &Self::Arena,
+    ) -> PolarsResult<VisitRecursion> {
+        let mut identifier = Identifier::new();
+        while let Some(Some(arg)) = self.arg_stack.pop() {
+            identifier.combine(&arg);
+        }
+        identifier = identifier.add_ae_node(node, arena);
+        let uniq_id = *self.identifier_to_uniq_id.entry(
+            identifier,
+            || {
+                let uniq_id = self.uniq_id_to_node.len() as u32;
+                self.uniq_id_to_node.push(node.node());
+                uniq_id
+            },
+            arena,
+        );
+        self.node_to_uniq_id.insert(node.node(), uniq_id);
+        Ok(VisitRecursion::Continue)
+    }
+}
+
 /// Identifier maps to Expr Node and count.
 type SubExprCount = IdentifierMap<(Node, u32)>;
 /// (post_visit_idx, identifier);
@@ -182,10 +246,6 @@ fn skip_pre_visit(ae: &AExpr, is_groupby: bool) -> bool {
     match ae {
         AExpr::Window { .. } => true,
         #[cfg(feature = "dtype-struct")]
-        AExpr::Function {
-            function: FunctionExpr::AsStruct,
-            ..
-        } => true,
         AExpr::Ternary { .. } => is_groupby,
         _ => false,
     }
@@ -308,7 +368,7 @@ impl ExprIdentifierVisitor<'_> {
             AExpr::Window { .. } => REFUSE_SKIP,
             // Don't allow this for now, as we can get `null().cast()` in ternary expressions.
             // TODO! Add a typed null
-            AExpr::Literal(LiteralValue::Null) => REFUSE_NO_MEMBER,
+            AExpr::Literal(LiteralValue::Scalar(sc)) if sc.is_null() => REFUSE_NO_MEMBER,
             AExpr::Literal(s) => {
                 match s {
                     LiteralValue::Series(s) => {
@@ -317,7 +377,7 @@ impl ExprIdentifierVisitor<'_> {
                         // Object and nested types are harder to hash and compare.
                         let allow = !(dtype.is_nested() | dtype.is_object());
 
-                        if s.len() < SERIES_LIMIT && allow {
+                        if s.len() < CHEAP_SERIES_HASH_LIMIT && allow {
                             REFUSE_ALLOW_MEMBER
                         } else {
                             REFUSE_NO_MEMBER
@@ -701,7 +761,9 @@ impl CommonSubExprOptimizer {
 
             if !valid {
                 if verbose() {
-                    eprintln!("materialized names collided in common subexpression elimination.\n backtrace and run without CSE")
+                    eprintln!(
+                        "materialized names collided in common subexpression elimination.\n backtrace and run without CSE"
+                    )
                 }
                 return Ok(None);
             }
@@ -724,6 +786,67 @@ impl CommonSubExprOptimizer {
                 } else {
                     out_e.set_node(out_node);
 
+                    // Ensure the function ExprIR's have the proper names.
+                    // This is needed for structs to get the proper field
+                    let mut scratch = vec![];
+                    let mut stack = vec![(e.node(), out_node)];
+                    while let Some((original, new)) = stack.pop() {
+                        // Don't follow identical nodes.
+                        if original == new {
+                            continue;
+                        }
+                        scratch.clear();
+                        let aes = expr_arena.get_many_mut([original, new]);
+
+                        // Only follow paths that are the same.
+                        if std::mem::discriminant(aes[0]) != std::mem::discriminant(aes[1]) {
+                            continue;
+                        }
+
+                        aes[0].inputs_rev(&mut scratch);
+                        let offset = scratch.len();
+                        aes[1].inputs_rev(&mut scratch);
+
+                        // If they have a different number of inputs, we don't follow the nodes.
+                        if scratch.len() != offset * 2 {
+                            continue;
+                        }
+
+                        for i in 0..scratch.len() / 2 {
+                            stack.push((scratch[i], scratch[i + offset]));
+                        }
+
+                        match expr_arena.get_many_mut([original, new]) {
+                            [
+                                AExpr::Function {
+                                    input: input_original,
+                                    ..
+                                },
+                                AExpr::Function {
+                                    input: input_new, ..
+                                },
+                            ] => {
+                                for (new, original) in input_new.iter_mut().zip(input_original) {
+                                    new.set_alias(original.output_name().clone());
+                                }
+                            },
+                            [
+                                AExpr::AnonymousFunction {
+                                    input: input_original,
+                                    ..
+                                },
+                                AExpr::AnonymousFunction {
+                                    input: input_new, ..
+                                },
+                            ] => {
+                                for (new, original) in input_new.iter_mut().zip(input_original) {
+                                    new.set_alias(original.output_name().clone());
+                                }
+                            },
+                            _ => {},
+                        }
+                    }
+
                     // If we don't end with an alias we add an alias. Because the normal left-hand
                     // rule we apply for determining the name will not work we now refer to
                     // intermediate temporary names starting with the `CSE_REPLACED` constant.
