@@ -9,10 +9,10 @@ use polars_plan::dsl::function_expr::StructFunction;
 use polars_plan::prelude::*;
 use polars_utils::format_pl_smallstr;
 use sqlparser::ast::{
-    BinaryOperator, CreateTable, Distinct, ExcludeSelectItem, Expr as SQLExpr, FunctionArg,
-    GroupByExpr, Ident, JoinConstraint, JoinOperator, ObjectName, ObjectType, Offset, OrderBy,
-    Query, RenameSelectItem, Select, SelectItem, SetExpr, SetOperator, SetQuantifier, Statement,
-    TableAlias, TableFactor, TableWithJoins, UnaryOperator, Value as SQLValue, Values,
+    BinaryOperator, CreateTable, Delete, Distinct, ExcludeSelectItem, Expr as SQLExpr, FromTable,
+    FunctionArg, GroupByExpr, Ident, JoinConstraint, JoinOperator, ObjectName, ObjectType, Offset,
+    OrderBy, Query, RenameSelectItem, Select, SelectItem, SetExpr, SetOperator, SetQuantifier,
+    Statement, TableAlias, TableFactor, TableWithJoins, UnaryOperator, Value as SQLValue, Values,
     WildcardAdditionalOptions,
 };
 use sqlparser::dialect::GenericDialect;
@@ -27,7 +27,7 @@ use crate::table_functions::PolarsTableFunctions;
 #[derive(Clone)]
 pub struct TableInfo {
     pub(crate) frame: LazyFrame,
-    pub(crate) name: String,
+    pub(crate) name: PlSmallStr,
     pub(crate) schema: Arc<Schema>,
 }
 
@@ -201,8 +201,9 @@ impl SQLContext {
             } => self.execute_drop_table(stmt)?,
             stmt @ Statement::Explain { .. } => self.execute_explain(stmt)?,
             stmt @ Statement::Truncate { .. } => self.execute_truncate_table(stmt)?,
+            stmt @ Statement::Delete { .. } => self.execute_delete_from_table(stmt)?,
             _ => polars_bail!(
-                SQLInterface: "statement type {:?} is not supported", ast,
+                SQLInterface: "statement type is not supported:\n{:?}", ast,
             ),
         })
     }
@@ -346,10 +347,6 @@ impl SQLContext {
                 }
             },
             op => {
-                let op = match op {
-                    SetExpr::SetOperation { op, .. } => op,
-                    _ => unreachable!(),
-                };
                 polars_bail!(SQLInterface: "'{}' operation is currently unsupported", op)
             },
         }
@@ -475,7 +472,7 @@ impl SQLContext {
                 let df = DataFrame::new(vec![plan])?;
                 Ok(df.lazy())
             },
-            _ => unreachable!(),
+            _ => polars_bail!(SQLInterface: "unexpected statement type; expected EXPLAIN"),
         }
     }
 
@@ -486,6 +483,7 @@ impl SQLContext {
         Ok(df.lazy())
     }
 
+    // DROP TABLE <tbl>
     fn execute_drop_table(&mut self, stmt: &Statement) -> PolarsResult<LazyFrame> {
         match stmt {
             Statement::Drop { names, .. } => {
@@ -494,10 +492,68 @@ impl SQLContext {
                 });
                 Ok(DataFrame::empty().lazy())
             },
-            _ => unreachable!(),
+            _ => polars_bail!(SQLInterface: "unexpected statement type; expected DROP"),
         }
     }
 
+    // DELETE FROM <tbl> [WHERE ...]
+    fn execute_delete_from_table(&mut self, stmt: &Statement) -> PolarsResult<LazyFrame> {
+        if let Statement::Delete(Delete {
+            tables,
+            from,
+            using,
+            selection,
+            returning,
+            order_by,
+            limit,
+        }) = stmt
+        {
+            if !tables.is_empty()
+                || using.is_some()
+                || returning.is_some()
+                || limit.is_some()
+                || !order_by.is_empty()
+            {
+                let error_message = match () {
+                    _ if !tables.is_empty() => "DELETE expects exactly one table name",
+                    _ if using.is_some() => "DELETE does not support the USING clause",
+                    _ if returning.is_some() => "DELETE does not support the RETURNING clause",
+                    _ if limit.is_some() => "DELETE does not support the LIMIT clause",
+                    _ if !order_by.is_empty() => "DELETE does not support the ORDER BY clause",
+                    _ => unreachable!(),
+                };
+                polars_bail!(SQLInterface: error_message);
+            }
+            let from_tables = match &from {
+                FromTable::WithFromKeyword(from) => from,
+                FromTable::WithoutKeyword(from) => from,
+            };
+            if from_tables.len() > 1 {
+                polars_bail!(SQLInterface: "cannot have multiple tables in DELETE FROM (found {})", from_tables.len())
+            }
+            let tbl_expr = from_tables.first().unwrap();
+            if !tbl_expr.joins.is_empty() {
+                polars_bail!(SQLInterface: "DELETE does not support table JOINs")
+            }
+            let (_, mut lf) = self.get_table(&tbl_expr.relation)?;
+            if selection.is_none() {
+                // no WHERE clause; equivalent to TRUNCATE (drop all rows)
+                Ok(DataFrame::empty_with_schema(
+                    lf.schema_with_arenas(&mut self.lp_arena, &mut self.expr_arena)
+                        .unwrap()
+                        .as_ref(),
+                )
+                .lazy())
+            } else {
+                // apply constraint as inverted filter (drops rows matching the selection)
+                Ok(self.process_where(lf.clone(), selection, true)?)
+            }
+        } else {
+            polars_bail!(SQLInterface: "unexpected statement type; expected DELETE")
+        }
+    }
+
+    // TRUNCATE <tbl>
     fn execute_truncate_table(&mut self, stmt: &Statement) -> PolarsResult<LazyFrame> {
         if let Statement::Truncate {
             table_names,
@@ -528,7 +584,7 @@ impl SQLContext {
                 },
             }
         } else {
-            unreachable!()
+            polars_bail!(SQLInterface: "unexpected statement type; expected TRUNCATE")
         }
     }
 
@@ -557,6 +613,13 @@ impl SQLContext {
         if !tbl_expr.joins.is_empty() {
             for join in &tbl_expr.joins {
                 let (r_name, mut rf) = self.get_table(&join.relation)?;
+                if r_name.is_empty() {
+                    // Require non-empty to avoid duplicate column errors from nested self-joins.
+                    polars_bail!(
+                        SQLInterface:
+                        "cannot join on unnamed relation; please provide an alias"
+                    )
+                }
                 let left_schema = self.get_frame_schema(&mut lf)?;
                 let right_schema = self.get_frame_schema(&mut rf)?;
 
@@ -578,12 +641,12 @@ impl SQLContext {
                         self.process_join(
                             &TableInfo {
                                 frame: lf,
-                                name: l_name.clone(),
+                                name: (&l_name).into(),
                                 schema: left_schema.clone(),
                             },
                             &TableInfo {
                                 frame: rf,
-                                name: r_name.clone(),
+                                name: (&r_name).into(),
                                 schema: right_schema.clone(),
                             },
                             constraint,
@@ -593,10 +656,18 @@ impl SQLContext {
                                 JoinOperator::RightOuter(_) => JoinType::Right,
                                 JoinOperator::Inner(_) => JoinType::Inner,
                                 #[cfg(feature = "semi_anti_join")]
-                                JoinOperator::Anti(_) | JoinOperator::LeftAnti(_) | JoinOperator::RightAnti(_) => JoinType::Anti,
+                                JoinOperator::Anti(_)
+                                | JoinOperator::LeftAnti(_)
+                                | JoinOperator::RightAnti(_) => JoinType::Anti,
                                 #[cfg(feature = "semi_anti_join")]
-                                JoinOperator::Semi(_) | JoinOperator::LeftSemi(_) | JoinOperator::RightSemi(_) => JoinType::Semi,
-                                join_type => polars_bail!(SQLInterface: "join type '{:?}' not currently supported", join_type),
+                                JoinOperator::Semi(_)
+                                | JoinOperator::LeftSemi(_)
+                                | JoinOperator::RightSemi(_) => JoinType::Semi,
+                                join_type => polars_bail!(
+                                    SQLInterface:
+                                    "join type '{:?}' not currently supported",
+                                    join_type
+                                ),
                             },
                         )?
                     },
@@ -612,7 +683,7 @@ impl SQLContext {
                 let joined_schema = self.get_frame_schema(&mut lf)?;
 
                 self.joined_aliases.borrow_mut().insert(
-                    r_name.to_string(),
+                    r_name.clone(),
                     right_schema
                         .iter_names()
                         .filter_map(|name| {
@@ -649,7 +720,7 @@ impl SQLContext {
 
         // Filter expression (WHERE clause)
         let schema = self.get_frame_schema(&mut lf)?;
-        lf = self.process_where(lf, &select_stmt.selection)?;
+        lf = self.process_where(lf, &select_stmt.selection, false)?;
 
         // 'SELECT *' modifiers
         let mut select_modifiers = SelectModifiers {
@@ -724,7 +795,8 @@ impl SQLContext {
             let mut retained_cols = Vec::with_capacity(projections.len());
             let mut retained_names = Vec::with_capacity(projections.len());
             let have_order_by = query.order_by.is_some();
-            let mut all_literal = true;
+            // Initialize containing InheritsContext to handle empty projection case.
+            let mut projection_heights = ExprSqlProjectionHeightBehavior::InheritsContext;
 
             // Note: if there is an 'order by' then we project everything (original cols
             // and new projections) and *then* select the final cols; the retained cols
@@ -738,7 +810,8 @@ impl SQLContext {
                 if select_modifiers.matches_ilike(&name)
                     && !select_modifiers.exclude.contains(&name)
                 {
-                    all_literal &= expr_to_leaf_column_names_iter(p).next().is_none();
+                    projection_heights |= ExprSqlProjectionHeightBehavior::identify_from_expr(p);
+
                     retained_cols.push(if have_order_by {
                         col(name.as_str())
                     } else {
@@ -750,8 +823,46 @@ impl SQLContext {
 
             // Apply the remaining modifiers and establish the final projection
             if have_order_by {
-                lf = lf.with_columns(projections);
+                // We can safely use `with_columns()` and avoid a join if:
+                // * There is already a projection that projects to the table height.
+                // * All projection heights inherit from context (e.g. all scalar literals that
+                //   are to be broadcasted to table height).
+                if projection_heights.contains(ExprSqlProjectionHeightBehavior::MaintainsColumn)
+                    || projection_heights == ExprSqlProjectionHeightBehavior::InheritsContext
+                {
+                    lf = lf.with_columns(projections);
+                } else {
+                    // We hit this branch if the output height is not guaranteed to match the table
+                    // height. E.g.:
+                    //
+                    // * SELECT COUNT(*) FROM df ORDER BY sort_key;
+                    // * SELECT UNNEST(list_col) FROM df ORDER BY sort_key;
+                    //
+                    // For these cases we truncate / extend the sorting columns with NULLs to match
+                    // the output height. We do this by projecting independently and then joining
+                    // back the original frame on the row index.
+                    const NAME: PlSmallStr = PlSmallStr::from_static("__PL_INDEX");
+                    lf = lf
+                        .clone()
+                        .select(projections)
+                        .with_row_index(NAME, None)
+                        .join(
+                            lf.with_row_index(NAME, None),
+                            [col(NAME)],
+                            [col(NAME)],
+                            JoinArgs {
+                                how: JoinType::Left,
+                                validation: Default::default(),
+                                suffix: None,
+                                slice: None,
+                                nulls_equal: false,
+                                coalesce: Default::default(),
+                                maintain_order: polars_ops::frame::MaintainOrderJoin::Left,
+                            },
+                        );
+                }
             }
+
             if !select_modifiers.replace.is_empty() {
                 lf = lf.with_columns(&select_modifiers.replace);
             }
@@ -761,7 +872,11 @@ impl SQLContext {
 
             lf = self.process_order_by(lf, &query.order_by, Some(&retained_cols))?;
 
-            if all_literal && !have_order_by {
+            // Note: If `have_order_by`, with_columns is already done above.
+            if projection_heights == ExprSqlProjectionHeightBehavior::InheritsContext
+                && !have_order_by
+            {
+                // All projections need to be broadcasted to table height, so evaluate in `with_columns()`
                 lf = lf.with_columns(retained_cols).select(retained_names);
             } else {
                 lf = lf.select(retained_cols);
@@ -867,6 +982,7 @@ impl SQLContext {
         &mut self,
         mut lf: LazyFrame,
         expr: &Option<SQLExpr>,
+        invert_filter: bool,
     ) -> PolarsResult<LazyFrame> {
         if let Some(expr) = expr {
             let schema = self.get_frame_schema(&mut lf)?;
@@ -883,9 +999,9 @@ impl SQLContext {
                 },
                 _ => (false, false),
             };
-            if all_true {
+            if (all_true && !invert_filter) || (all_false && invert_filter) {
                 return Ok(lf);
-            } else if all_false {
+            } else if (all_false && !invert_filter) || (all_true && invert_filter) {
                 return Ok(DataFrame::empty_with_schema(schema.as_ref()).lazy());
             }
 
@@ -895,7 +1011,11 @@ impl SQLContext {
                 filter_expression = all_horizontal([filter_expression])?;
             }
             lf = self.process_subqueries(lf, vec![&mut filter_expression]);
-            lf = lf.filter(filter_expression);
+            lf = if invert_filter {
+                lf.remove(filter_expression)
+            } else {
+                lf.filter(filter_expression)
+            };
         }
         Ok(lf)
     }
@@ -1024,7 +1144,6 @@ impl SQLContext {
                 ..
             } => {
                 if let Some(alias) = alias {
-                    let table_name = alias.name.value.clone();
                     let column_names: Vec<Option<PlSmallStr>> = alias
                         .columns
                         .iter()
@@ -1067,11 +1186,13 @@ impl SQLContext {
                         .collect();
 
                     let lf = DataFrame::new(column_series)?.lazy();
+
                     if *with_offset {
-                        // TODO: support 'WITH ORDINALITY' modifier.
+                        // TODO: support 'WITH ORDINALITY|OFFSET' modifier.
                         //  (note that 'WITH OFFSET' is BigQuery-specific syntax, not PostgreSQL)
-                        polars_bail!(SQLInterface: "UNNEST tables do not (yet) support WITH OFFSET/ORDINALITY");
+                        polars_bail!(SQLInterface: "UNNEST tables do not (yet) support WITH ORDINALITY|OFFSET");
                     }
+                    let table_name = alias.name.value.clone();
                     self.table_map.insert(table_name.clone(), lf.clone());
                     Ok((table_name.clone(), lf))
                 } else {
@@ -1354,7 +1475,10 @@ impl SQLContext {
                 .replace('%', ".*")
                 .replace('_', ".");
 
-            modifiers.ilike = Some(regex::Regex::new(format!("^(?is){}$", rx).as_str()).unwrap());
+            modifiers.ilike = Some(
+                polars_utils::regex_cache::compile_regex(format!("^(?is){}$", rx).as_str())
+                    .unwrap(),
+            );
         }
 
         // SELECT * RENAME
@@ -1451,10 +1575,10 @@ fn expand_exprs(expr: Expr, schema: &SchemaRef) -> Vec<Expr> {
             .map(|name| col(name.clone()))
             .collect::<Vec<_>>(),
         Expr::Column(nm) if is_regex_colname(nm.as_str()) => {
-            let rx = regex::Regex::new(&nm).unwrap();
+            let re = polars_utils::regex_cache::compile_regex(&nm).unwrap();
             schema
                 .iter_names()
-                .filter(|name| rx.is_match(name))
+                .filter(|name| re.is_match(name))
                 .map(|name| col(name.clone()))
                 .collect::<Vec<_>>()
         },
@@ -1531,3 +1655,57 @@ fn process_join_constraint(
         _ => polars_bail!(SQLInterface: "unsupported SQL join constraint:\n{:?}", constraint),
     }
 }
+
+bitflags::bitflags! {
+    /// Bitfield indicating whether there exists a projection with the specified height behavior.
+    ///
+    /// Used to help determine whether to execute projections in `select()` or `with_columns()`
+    /// context.
+    #[derive(PartialEq)]
+    struct ExprSqlProjectionHeightBehavior: u8 {
+        /// Maintains the height of input column(s)
+        const MaintainsColumn = 1 << 0;
+        /// Height is independent of input, e.g.:
+        /// * expressions that change length: e.g. slice, explode, filter, gather etc.
+        /// * aggregations: count(*), first(), sum() etc.
+        const Independent = 1 << 1;
+        /// "Inherits" the height of the context, e.g.:
+        /// * Scalar literals
+        const InheritsContext = 1 << 2;
+    }
+}
+
+impl ExprSqlProjectionHeightBehavior {
+    fn identify_from_expr(expr: &Expr) -> Self {
+        let mut has_column = false;
+        let mut has_independent = false;
+
+        for e in expr.into_iter() {
+            use Expr::*;
+
+            has_column |= matches!(e, Column(_) | Columns(_) | DtypeColumn(_) | IndexColumn(_));
+
+            has_independent |= match e {
+                Function { options, .. } | AnonymousFunction { options, .. } => {
+                    options.returns_scalar() || !options.is_length_preserving()
+                },
+
+                Literal(v) => !v.is_scalar(),
+
+                Explode { .. } | Filter { .. } | Gather { .. } | Slice { .. } => true,
+
+                Agg { .. } | Len => true,
+
+                _ => false,
+            }
+        }
+
+        if has_independent {
+            Self::Independent
+        } else if has_column {
+            Self::MaintainsColumn
+        } else {
+            Self::InheritsContext
+        }
+    }
+}
