@@ -11,27 +11,26 @@ use std::ops::Div;
 
 use polars_core::prelude::*;
 use polars_lazy::prelude::*;
+use polars_plan::plans::DynLiteralValue;
 use polars_plan::prelude::typed_lit;
-use polars_plan::prelude::LiteralValue::Null;
 use polars_time::Duration;
 use rand::distributions::Alphanumeric;
-use rand::{thread_rng, Rng};
+use rand::{Rng, thread_rng};
 #[cfg(feature = "serde")]
 use serde::{Deserialize, Serialize};
 use sqlparser::ast::{
-    BinaryOperator as SQLBinaryOperator, BinaryOperator, CastFormat, CastKind,
-    DataType as SQLDataType, DateTimeField, Expr as SQLExpr, Function as SQLFunction, Ident,
-    Interval, Query as Subquery, SelectItem, Subscript, TimezoneInfo, TrimWhereField,
-    UnaryOperator, Value as SQLValue,
+    BinaryOperator as SQLBinaryOperator, CastFormat, CastKind, DataType as SQLDataType,
+    DateTimeField, Expr as SQLExpr, Function as SQLFunction, Ident, Interval, Query as Subquery,
+    SelectItem, Subscript, TimezoneInfo, TrimWhereField, UnaryOperator, Value as SQLValue,
 };
 use sqlparser::dialect::GenericDialect;
 use sqlparser::parser::{Parser, ParserOptions};
 
+use crate::SQLContext;
 use crate::functions::SQLFunctionVisitor;
 use crate::types::{
     bitstring_to_bytes_literal, is_iso_date, is_iso_datetime, is_iso_time, map_sql_dtype_to_polars,
 };
-use crate::SQLContext;
 
 #[inline]
 #[cold]
@@ -60,24 +59,23 @@ pub(crate) struct SQLExprVisitor<'a> {
 
 impl SQLExprVisitor<'_> {
     fn array_expr_to_series(&mut self, elements: &[SQLExpr]) -> PolarsResult<Series> {
-        let array_elements = elements
-            .iter()
-            .map(|e| match e {
+        let mut array_elements = Vec::with_capacity(elements.len());
+        for e in elements {
+            let val = match e {
                 SQLExpr::Value(v) => self.visit_any_value(v, None),
                 SQLExpr::UnaryOp { op, expr } => match expr.as_ref() {
                     SQLExpr::Value(v) => self.visit_any_value(v, Some(op)),
-                    _ => Err(polars_err!(SQLInterface: "expression {:?} is not currently supported", e)),
+                    _ => Err(polars_err!(SQLInterface: "array element {:?} is not supported", e)),
                 },
-                SQLExpr::Array(_) => {
-                    // TODO: nested arrays (handle FnMut issues)
-                    // let srs = self.array_expr_to_series(&[e.clone()])?;
-                    // Ok(AnyValue::List(srs))
-                    Err(polars_err!(SQLInterface: "nested array literals are not currently supported:\n{:?}", e))
+                SQLExpr::Array(values) => {
+                    let srs = self.array_expr_to_series(&values.elem)?;
+                    Ok(AnyValue::List(srs))
                 },
-                _ => Err(polars_err!(SQLInterface: "expression {:?} is not currently supported", e)),
-            })
-            .collect::<PolarsResult<Vec<_>>>()?;
-
+                _ => Err(polars_err!(SQLInterface: "array element {:?} is not supported", e)),
+            }?
+            .into_static();
+            array_elements.push(val);
+        }
         Series::from_any_values(PlSmallStr::EMPTY, &array_elements, true)
     }
 
@@ -124,8 +122,8 @@ impl SQLExprVisitor<'_> {
                 negated,
             } => {
                 let expr = self.visit_expr(expr)?;
-                let elems = self.visit_array_expr(list, false, Some(&expr))?;
-                let is_in = expr.is_in(elems);
+                let elems = self.visit_array_expr(list, true, Some(&expr))?;
+                let is_in = expr.is_in(elems, false);
                 Ok(if *negated { is_in.not() } else { is_in })
             },
             SQLExpr::InSubquery {
@@ -133,7 +131,7 @@ impl SQLExprVisitor<'_> {
                 subquery,
                 negated,
             } => self.visit_in_subquery(expr, subquery, *negated),
-            SQLExpr::Interval(interval) => self.visit_interval(interval),
+            SQLExpr::Interval(interval) => Ok(lit(interval_to_duration(interval, true)?)),
             SQLExpr::IsDistinctFrom(e1, e2) => {
                 Ok(self.visit_expr(e1)?.neq_missing(self.visit_expr(e2)?))
             },
@@ -301,30 +299,6 @@ impl SQLExprVisitor<'_> {
         Ok(resolve_compound_identifier(self.ctx, idents, self.active_schema)?[0].clone())
     }
 
-    fn visit_interval(&self, interval: &Interval) -> PolarsResult<Expr> {
-        if interval.last_field.is_some()
-            || interval.leading_field.is_some()
-            || interval.leading_precision.is_some()
-            || interval.fractional_seconds_precision.is_some()
-        {
-            polars_bail!(SQLSyntax: "unsupported interval syntax ('{}')", interval)
-        }
-        let s = match &*interval.value {
-            SQLExpr::UnaryOp { .. } => {
-                polars_bail!(SQLSyntax: "unary ops are not valid on interval strings; found {}", interval.value)
-            },
-            SQLExpr::Value(SQLValue::SingleQuotedString(s)) => Some(s),
-            _ => None,
-        };
-        match s {
-            Some(s) if s.contains('-') => {
-                polars_bail!(SQLInterface: "minus signs are not yet supported in interval strings; found '{}'", s)
-            },
-            Some(s) => Ok(lit(Duration::parse_interval(s))),
-            None => polars_bail!(SQLSyntax: "invalid interval {:?}", interval),
-        }
-    }
-
     fn visit_like(
         &mut self,
         negated: bool,
@@ -337,7 +311,9 @@ impl SQLExprVisitor<'_> {
             polars_bail!(SQLInterface: "ESCAPE char for LIKE/ILIKE is not currently supported; found '{}'", escape_char.clone().unwrap());
         }
         let pat = match self.visit_expr(pattern) {
-            Ok(Expr::Literal(LiteralValue::String(s))) => s,
+            Ok(Expr::Literal(lv)) if lv.extract_str().is_some() => {
+                PlSmallStr::from_str(lv.extract_str().unwrap())
+            },
             _ => {
                 polars_bail!(SQLSyntax: "LIKE/ILIKE pattern must be a string literal; found {}", pattern)
             },
@@ -345,9 +321,9 @@ impl SQLExprVisitor<'_> {
         if pat.is_empty() || (!case_insensitive && pat.chars().all(|c| !matches!(c, '%' | '_'))) {
             // empty string or other exact literal match (eg: no wildcard chars)
             let op = if negated {
-                BinaryOperator::NotEq
+                SQLBinaryOperator::NotEq
             } else {
-                BinaryOperator::Eq
+                SQLBinaryOperator::Eq
             };
             self.visit_binary_op(expr, &op, pattern)
         } else {
@@ -383,15 +359,19 @@ impl SQLExprVisitor<'_> {
 
     /// Handle implicit temporal string comparisons.
     ///
-    /// eg: "dt >= '2024-04-30'", or "dtm::date = '2077-10-10'"
+    /// eg: clauses such as -
+    ///   "dt >= '2024-04-30'"
+    ///   "dt = '2077-10-10'::date"
+    ///   "dtm::date = '2077-10-10'
     fn convert_temporal_strings(&mut self, left: &Expr, right: &Expr) -> Expr {
         if let (Some(name), Some(s), expr_dtype) = match (left, right) {
             // identify "col <op> string" expressions
-            (Expr::Column(name), Expr::Literal(LiteralValue::String(s))) => {
-                (Some(name.clone()), Some(s), None)
+            (Expr::Column(name), Expr::Literal(lv)) if lv.extract_str().is_some() => {
+                (Some(name.clone()), Some(lv.extract_str().unwrap()), None)
             },
             // identify "CAST(expr AS type) <op> string" and/or "expr::type <op> string" expressions
-            (Expr::Cast { expr, dtype, .. }, Expr::Literal(LiteralValue::String(s))) => {
+            (Expr::Cast { expr, dtype, .. }, Expr::Literal(lv)) if lv.extract_str().is_some() => {
+                let s = lv.extract_str().unwrap();
                 match &**expr {
                     Expr::Column(name) => (Some(name.clone()), Some(s), Some(dtype)),
                     _ => (None, Some(s), Some(dtype)),
@@ -480,11 +460,45 @@ impl SQLExprVisitor<'_> {
     fn visit_binary_op(
         &mut self,
         left: &SQLExpr,
-        op: &BinaryOperator,
+        op: &SQLBinaryOperator,
         right: &SQLExpr,
     ) -> PolarsResult<Expr> {
-        let lhs = self.visit_expr(left)?;
-        let mut rhs = self.visit_expr(right)?;
+        // need special handling for interval offsets and comparisons
+        let (lhs, mut rhs) = match (left, op, right) {
+            (_, SQLBinaryOperator::Minus, SQLExpr::Interval(v)) => {
+                let duration = interval_to_duration(v, false)?;
+                return Ok(self
+                    .visit_expr(left)?
+                    .dt()
+                    .offset_by(lit(format!("-{}", duration))));
+            },
+            (_, SQLBinaryOperator::Plus, SQLExpr::Interval(v)) => {
+                let duration = interval_to_duration(v, false)?;
+                return Ok(self
+                    .visit_expr(left)?
+                    .dt()
+                    .offset_by(lit(format!("{}", duration))));
+            },
+            (SQLExpr::Interval(v1), _, SQLExpr::Interval(v2)) => {
+                // shortcut interval comparison evaluation (-> bool)
+                let d1 = interval_to_duration(v1, false)?;
+                let d2 = interval_to_duration(v2, false)?;
+                let res = match op {
+                    SQLBinaryOperator::Gt => Ok(lit(d1 > d2)),
+                    SQLBinaryOperator::Lt => Ok(lit(d1 < d2)),
+                    SQLBinaryOperator::GtEq => Ok(lit(d1 >= d2)),
+                    SQLBinaryOperator::LtEq => Ok(lit(d1 <= d2)),
+                    SQLBinaryOperator::NotEq => Ok(lit(d1 != d2)),
+                    SQLBinaryOperator::Eq | SQLBinaryOperator::Spaceship => Ok(lit(d1 == d2)),
+                    _ => polars_bail!(SQLInterface: "invalid interval comparison operator"),
+                };
+                if res.is_ok() {
+                    return res;
+                }
+                (self.visit_expr(left)?, self.visit_expr(right)?)
+            },
+            _ => (self.visit_expr(left)?, self.visit_expr(right)?),
+        };
         rhs = self.convert_temporal_strings(&lhs, &rhs);
 
         Ok(match op {
@@ -521,21 +535,23 @@ impl SQLExprVisitor<'_> {
             // Regular expression operators
             // ----
             SQLBinaryOperator::PGRegexMatch => match rhs {  // "x ~ y"
-                Expr::Literal(LiteralValue::String(_)) => lhs.str().contains(rhs, true),
+                Expr::Literal(ref lv) if lv.extract_str().is_some() => lhs.str().contains(rhs, true),
                 _ => polars_bail!(SQLSyntax: "invalid pattern for '~' operator: {:?}", rhs),
             },
             SQLBinaryOperator::PGRegexNotMatch => match rhs {  // "x !~ y"
-                Expr::Literal(LiteralValue::String(_)) => lhs.str().contains(rhs, true).not(),
+                Expr::Literal(ref lv) if lv.extract_str().is_some() => lhs.str().contains(rhs, true).not(),
                 _ => polars_bail!(SQLSyntax: "invalid pattern for '!~' operator: {:?}", rhs),
             },
             SQLBinaryOperator::PGRegexIMatch => match rhs {  // "x ~* y"
-                Expr::Literal(LiteralValue::String(pat)) => {
+                Expr::Literal(ref lv) if lv.extract_str().is_some() => {
+                    let pat = lv.extract_str().unwrap();
                     lhs.str().contains(lit(format!("(?i){}", pat)), true)
                 },
                 _ => polars_bail!(SQLSyntax: "invalid pattern for '~*' operator: {:?}", rhs),
             },
             SQLBinaryOperator::PGRegexNotIMatch => match rhs {  // "x !~* y"
-                Expr::Literal(LiteralValue::String(pat)) => {
+                Expr::Literal(ref lv) if lv.extract_str().is_some() => {
+                    let pat = lv.extract_str().unwrap();
                     lhs.str().contains(lit(format!("(?i){}", pat)), true).not()
                 },
                 _ => {
@@ -575,14 +591,15 @@ impl SQLExprVisitor<'_> {
             // JSON/Struct field access operators
             // ----
             SQLBinaryOperator::Arrow | SQLBinaryOperator::LongArrow => match rhs {  // "x -> y", "x ->> y"
-                Expr::Literal(LiteralValue::String(path)) => {
-                    let mut expr = self.struct_field_access_expr(&lhs, &path, false)?;
+                Expr::Literal(lv) if lv.extract_str().is_some() => {
+                    let path = lv.extract_str().unwrap();
+                    let mut expr = self.struct_field_access_expr(&lhs, path, false)?;
                     if let SQLBinaryOperator::LongArrow = op {
                         expr = expr.cast(DataType::String);
                     }
                     expr
                 },
-                Expr::Literal(LiteralValue::Int(idx)) => {
+                Expr::Literal(LiteralValue::Dyn(DynLiteralValue::Int(idx))) => {
                     let mut expr = self.struct_field_access_expr(&lhs, &idx.to_string(), true)?;
                     if let SQLBinaryOperator::LongArrow = op {
                         expr = expr.cast(DataType::String);
@@ -594,14 +611,18 @@ impl SQLExprVisitor<'_> {
                 },
             },
             SQLBinaryOperator::HashArrow | SQLBinaryOperator::HashLongArrow => {  // "x #> y", "x #>> y"
-                if let Expr::Literal(LiteralValue::String(path)) = rhs {
-                    let mut expr = self.struct_field_access_expr(&lhs, &path, true)?;
-                    if let SQLBinaryOperator::HashLongArrow = op {
-                        expr = expr.cast(DataType::String);
+                match rhs {
+                    Expr::Literal(lv) if lv.extract_str().is_some() => {
+                        let path = lv.extract_str().unwrap();
+                        let mut expr = self.struct_field_access_expr(&lhs, path, true)?;
+                        if let SQLBinaryOperator::HashLongArrow = op {
+                            expr = expr.cast(DataType::String);
+                        }
+                        expr
+                    },
+                    _ => {
+                        polars_bail!(SQLSyntax: "invalid json/struct path-extract definition: {:?}", rhs)
                     }
-                    expr
-                } else {
-                    polars_bail!(SQLSyntax: "invalid json/struct path-extract definition: {:?}", rhs)
                 }
             },
             other => {
@@ -617,10 +638,18 @@ impl SQLExprVisitor<'_> {
         let expr = self.visit_expr(expr)?;
         Ok(match (op, expr.clone()) {
             // simplify the parse tree by special-casing common unary +/- ops
-            (UnaryOperator::Plus, Expr::Literal(LiteralValue::Int(n))) => lit(n),
-            (UnaryOperator::Plus, Expr::Literal(LiteralValue::Float(n))) => lit(n),
-            (UnaryOperator::Minus, Expr::Literal(LiteralValue::Int(n))) => lit(-n),
-            (UnaryOperator::Minus, Expr::Literal(LiteralValue::Float(n))) => lit(-n),
+            (UnaryOperator::Plus, Expr::Literal(LiteralValue::Dyn(DynLiteralValue::Int(n)))) => {
+                lit(n)
+            },
+            (UnaryOperator::Plus, Expr::Literal(LiteralValue::Dyn(DynLiteralValue::Float(n)))) => {
+                lit(n)
+            },
+            (UnaryOperator::Minus, Expr::Literal(LiteralValue::Dyn(DynLiteralValue::Int(n)))) => {
+                lit(-n)
+            },
+            (UnaryOperator::Minus, Expr::Literal(LiteralValue::Dyn(DynLiteralValue::Float(n)))) => {
+                lit(-n)
+            },
             // general case
             (UnaryOperator::Plus, _) => lit(0) + expr,
             (UnaryOperator::Minus, _) => lit(0) - expr,
@@ -649,19 +678,19 @@ impl SQLExprVisitor<'_> {
     fn visit_all(
         &mut self,
         left: &SQLExpr,
-        compare_op: &BinaryOperator,
+        compare_op: &SQLBinaryOperator,
         right: &SQLExpr,
     ) -> PolarsResult<Expr> {
         let left = self.visit_expr(left)?;
         let right = self.visit_expr(right)?;
 
         match compare_op {
-            BinaryOperator::Gt => Ok(left.gt(right.max())),
-            BinaryOperator::Lt => Ok(left.lt(right.min())),
-            BinaryOperator::GtEq => Ok(left.gt_eq(right.max())),
-            BinaryOperator::LtEq => Ok(left.lt_eq(right.min())),
-            BinaryOperator::Eq => polars_bail!(SQLSyntax: "ALL cannot be used with ="),
-            BinaryOperator::NotEq => polars_bail!(SQLSyntax: "ALL cannot be used with !="),
+            SQLBinaryOperator::Gt => Ok(left.gt(right.max())),
+            SQLBinaryOperator::Lt => Ok(left.lt(right.min())),
+            SQLBinaryOperator::GtEq => Ok(left.gt_eq(right.max())),
+            SQLBinaryOperator::LtEq => Ok(left.lt_eq(right.min())),
+            SQLBinaryOperator::Eq => polars_bail!(SQLSyntax: "ALL cannot be used with ="),
+            SQLBinaryOperator::NotEq => polars_bail!(SQLSyntax: "ALL cannot be used with !="),
             _ => polars_bail!(SQLInterface: "invalid comparison operator"),
         }
     }
@@ -672,19 +701,19 @@ impl SQLExprVisitor<'_> {
     fn visit_any(
         &mut self,
         left: &SQLExpr,
-        compare_op: &BinaryOperator,
+        compare_op: &SQLBinaryOperator,
         right: &SQLExpr,
     ) -> PolarsResult<Expr> {
         let left = self.visit_expr(left)?;
         let right = self.visit_expr(right)?;
 
         match compare_op {
-            BinaryOperator::Gt => Ok(left.gt(right.min())),
-            BinaryOperator::Lt => Ok(left.lt(right.max())),
-            BinaryOperator::GtEq => Ok(left.gt_eq(right.min())),
-            BinaryOperator::LtEq => Ok(left.lt_eq(right.max())),
-            BinaryOperator::Eq => Ok(left.is_in(right)),
-            BinaryOperator::NotEq => Ok(left.is_in(right).not()),
+            SQLBinaryOperator::Gt => Ok(left.gt(right.min())),
+            SQLBinaryOperator::Lt => Ok(left.lt(right.max())),
+            SQLBinaryOperator::GtEq => Ok(left.gt_eq(right.min())),
+            SQLBinaryOperator::LtEq => Ok(left.lt_eq(right.max())),
+            SQLBinaryOperator::Eq => Ok(left.is_in(right, false)),
+            SQLBinaryOperator::NotEq => Ok(left.is_in(right, false).not()),
             _ => polars_bail!(SQLInterface: "invalid comparison operator"),
         }
     }
@@ -770,7 +799,7 @@ impl SQLExprVisitor<'_> {
                 };
                 lit(hex::decode(x.clone()).unwrap())
             },
-            SQLValue::Null => Expr::Literal(LiteralValue::Null),
+            SQLValue::Null => Expr::Literal(LiteralValue::untyped_null()),
             SQLValue::Number(s, _) => {
                 // Check for existence of decimal separator dot
                 if s.contains('.') {
@@ -835,7 +864,9 @@ impl SQLExprVisitor<'_> {
                 // note: for PostgreSQL this represents a BIT literal (eg: b'10101') not BYTE
                 let bytes_literal = bitstring_to_bytes_literal(b)?;
                 match bytes_literal {
-                    Expr::Literal(LiteralValue::Binary(v)) => AnyValue::BinaryOwned(v.to_vec()),
+                    Expr::Literal(lv) if lv.extract_binary().is_some() => {
+                        AnyValue::BinaryOwned(lv.extract_binary().unwrap().to_vec())
+                    },
                     _ => {
                         polars_bail!(SQLInterface: "failed to parse bitstring literal: {:?}", b)
                     },
@@ -884,16 +915,24 @@ impl SQLExprVisitor<'_> {
         let expr = self.visit_expr(expr)?;
         let trim_what = trim_what.as_ref().map(|e| self.visit_expr(e)).transpose()?;
         let trim_what = match trim_what {
-            Some(Expr::Literal(LiteralValue::String(val))) => Some(val),
+            Some(Expr::Literal(lv)) if lv.extract_str().is_some() => {
+                Some(PlSmallStr::from_str(lv.extract_str().unwrap()))
+            },
             None => None,
             _ => return self.err(&expr),
         };
         Ok(match (trim_where, trim_what) {
-            (None | Some(TrimWhereField::Both), None) => expr.str().strip_chars(lit(Null)),
+            (None | Some(TrimWhereField::Both), None) => {
+                expr.str().strip_chars(lit(LiteralValue::untyped_null()))
+            },
             (None | Some(TrimWhereField::Both), Some(val)) => expr.str().strip_chars(lit(val)),
-            (Some(TrimWhereField::Leading), None) => expr.str().strip_chars_start(lit(Null)),
+            (Some(TrimWhereField::Leading), None) => expr
+                .str()
+                .strip_chars_start(lit(LiteralValue::untyped_null())),
             (Some(TrimWhereField::Leading), Some(val)) => expr.str().strip_chars_start(lit(val)),
-            (Some(TrimWhereField::Trailing), None) => expr.str().strip_chars_end(lit(Null)),
+            (Some(TrimWhereField::Trailing), None) => expr
+                .str()
+                .strip_chars_end(lit(LiteralValue::untyped_null())),
             (Some(TrimWhereField::Trailing), Some(val)) => expr.str().strip_chars_end(lit(val)),
         })
     }
@@ -908,9 +947,9 @@ impl SQLExprVisitor<'_> {
         let subquery_result = self.visit_subquery(subquery, SubqueryRestriction::SingleColumn)?;
         let expr = self.visit_expr(expr)?;
         Ok(if negated {
-            expr.is_in(subquery_result).not()
+            expr.is_in(subquery_result, false).not()
         } else {
-            expr.is_in(subquery_result)
+            expr.is_in(subquery_result, false)
         })
     }
 
@@ -939,7 +978,7 @@ impl SQLExprVisitor<'_> {
             }
             let else_res = match else_result {
                 Some(else_res) => self.visit_expr(else_res)?,
-                None => lit(Null), // ELSE clause is optional; when omitted, it is implicitly NULL
+                None => lit(LiteralValue::untyped_null()), // ELSE clause is optional; when omitted, it is implicitly NULL
             };
             if let Some(operand_expr) = operand {
                 let first_operand_expr = self.visit_expr(operand_expr)?;
@@ -1037,6 +1076,38 @@ pub fn sql_expr<S: AsRef<str>>(s: S) -> PolarsResult<Expr> {
     })
 }
 
+pub(crate) fn interval_to_duration(interval: &Interval, fixed: bool) -> PolarsResult<Duration> {
+    if interval.last_field.is_some()
+        || interval.leading_field.is_some()
+        || interval.leading_precision.is_some()
+        || interval.fractional_seconds_precision.is_some()
+    {
+        polars_bail!(SQLSyntax: "unsupported interval syntax ('{}')", interval)
+    }
+    let s = match &*interval.value {
+        SQLExpr::UnaryOp { .. } => {
+            polars_bail!(SQLSyntax: "unary ops are not valid on interval strings; found {}", interval.value)
+        },
+        SQLExpr::Value(SQLValue::SingleQuotedString(s)) => Some(s),
+        _ => None,
+    };
+    match s {
+        Some(s) if s.contains('-') => {
+            polars_bail!(SQLInterface: "minus signs are not yet supported in interval strings; found '{}'", s)
+        },
+        Some(s) => {
+            // years, quarters, and months do not have a fixed duration; these
+            // interval parts can only be used with respect to a reference point
+            let duration = Duration::parse_interval(s);
+            if fixed && duration.months() != 0 {
+                polars_bail!(SQLSyntax: "fixed-duration interval cannot contain years, quarters, or months; found {}", s)
+            };
+            Ok(duration)
+        },
+        None => polars_bail!(SQLSyntax: "invalid interval {:?}", interval),
+    }
+}
+
 pub(crate) fn parse_sql_expr(
     expr: &SQLExpr,
     ctx: &mut SQLContext,
@@ -1152,23 +1223,23 @@ pub(crate) fn parse_extract_date_part(expr: Expr, field: &DateTimeField) -> Pola
 /// be adjusted from 1-indexed (SQL) to 0-indexed (Rust/Polars)
 pub(crate) fn adjust_one_indexed_param(idx: Expr, null_if_zero: bool) -> Expr {
     match idx {
-        Expr::Literal(Null) => lit(Null),
-        Expr::Literal(LiteralValue::Int(0)) => {
+        Expr::Literal(sc) if sc.is_null() => lit(LiteralValue::untyped_null()),
+        Expr::Literal(LiteralValue::Dyn(DynLiteralValue::Int(0))) => {
             if null_if_zero {
-                lit(Null)
+                lit(LiteralValue::untyped_null())
             } else {
                 idx
             }
         },
-        Expr::Literal(LiteralValue::Int(n)) if n < 0 => idx,
-        Expr::Literal(LiteralValue::Int(n)) => lit(n - 1),
+        Expr::Literal(LiteralValue::Dyn(DynLiteralValue::Int(n))) if n < 0 => idx,
+        Expr::Literal(LiteralValue::Dyn(DynLiteralValue::Int(n))) => lit(n - 1),
         // TODO: when 'saturating_sub' is available, should be able
         //  to streamline the when/then/otherwise block below -
         _ => when(idx.clone().gt(lit(0)))
             .then(idx.clone() - lit(1))
             .otherwise(if null_if_zero {
                 when(idx.clone().eq(lit(0)))
-                    .then(lit(Null))
+                    .then(lit(LiteralValue::untyped_null()))
                     .otherwise(idx.clone())
             } else {
                 idx.clone()
@@ -1223,14 +1294,21 @@ pub(crate) fn resolve_compound_identifier(
                 .iter_names_and_dtypes()
                 .map(|(name, dtype)| resolve_column(ctx, ident_root, name, dtype).unwrap().0)
                 .collect::<Vec<_>>());
-        } else if let Some((_, name, dtype)) = schema.get_full(name) {
-            resolve_column(ctx, ident_root, name, dtype)
-        } else if lf.is_none() {
+        };
+        let root_is_field = schema.get(&ident_root.value).is_some();
+        if lf.is_none() && root_is_field {
             remaining_idents = idents.iter().skip(1);
             Ok((
                 col(ident_root.value.as_str()),
                 schema.get(&ident_root.value),
             ))
+        } else if lf.is_none() && !root_is_field {
+            polars_bail!(
+                SQLInterface: "no table or struct column named '{}' found",
+                ident_root
+            )
+        } else if let Some((_, name, dtype)) = schema.get_full(name) {
+            resolve_column(ctx, ident_root, name, dtype)
         } else {
             polars_bail!(
                 SQLInterface: "no column named '{}' found in table '{}'",
@@ -1249,7 +1327,7 @@ pub(crate) fn resolve_compound_identifier(
                 return Ok(fields
                     .iter()
                     .map(|fld| column.clone().struct_().field_by_name(&fld.name))
-                    .collect())
+                    .collect());
             },
             Some(DataType::Struct(fields)) => {
                 dtype = fields
